import featureform as ff
from featureform.resources import Model
import os
import pytest
import time

real_path = os.path.realpath(__file__)
dir_path = os.path.dirname(real_path)


@pytest.mark.parametrize(
    "is_local,is_insecure",
    [
        pytest.param(False, False, marks=pytest.mark.hosted),
        pytest.param(False, True, marks=pytest.mark.docker),
    ],
)
def test_adding_tags_and_properties_to_provider(is_local, is_insecure, request):
    name = "postgres_docs"
    tags = ["pg_offline_provider", "ff_team"]
    properties = {"primary_offline_provider": "true"}

    ff.register_postgres(
        name=name,
        description="Example offline store store",
        team="Featureform",
        host="0.0.0.0",
        port="5432",
        user="postgres",
        password="password",
        database="postgres",
        tags=tags,
        properties=properties,
    )

    resource_client = ff.ResourceClient(local=is_local, insecure=is_insecure)
    resource_client.apply()

    postgres = resource_client.get_provider(name, is_local)
    assert (
        get_tags(postgres, is_local) == tags
        and get_properties(postgres, is_local) == properties
    )


@pytest.mark.parametrize(
    "is_local,is_insecure",
    [
        pytest.param(False, False, marks=pytest.mark.hosted),
        pytest.param(False, True, marks=pytest.mark.docker),
    ],
)
def test_updating_tags_and_properties_for_provider(is_local, is_insecure, request):
    name = "postgres_docs"
    tags = ["pg_offline_provider", "ff_team"]
    properties = {"primary_offline_provider": "true"}

    ff.register_postgres(
        name=name,
        description="Example offline store store",
        team="Featureform",
        host="0.0.0.0",
        port="5432",
        user="postgres",
        password="password",
        database="postgres",
        tags=tags,
        properties=properties,
    )

    resource_client = ff.ResourceClient(local=is_local, insecure=is_insecure)
    resource_client.apply()

    ff.clear_state()

    additional_tags = ["training_data"]
    additional_properties = {"is_active": "true"}

    ff.register_postgres(
        name=name,
        description="Example offline store store",
        team="Featureform",
        host="0.0.0.0",
        port="5432",
        user="postgres",
        password="password",
        database="postgres",
        tags=additional_tags,
        properties=additional_properties,
    )
    resource_client.apply()
    postgres = resource_client.get_provider(name, is_local)

    expected_tags = set(tags + additional_tags)
    expected_properties = {**properties, **additional_properties}

    actual_tags = set(get_tags(postgres, is_local))
    actual_properties = get_properties(postgres, is_local)

    assert actual_tags == expected_tags and actual_properties == expected_properties


@pytest.mark.parametrize(
    "is_local,is_insecure",
    [
        pytest.param(False, False, marks=pytest.mark.hosted),
        pytest.param(False, True, marks=pytest.mark.docker),
    ],
)
def test_adding_tags_and_properties_to_user(is_local, is_insecure, request):
    username = "ff_user"
    tags = ["primary_user"]
    properties = {"rotated_credentials": "yes"}

    ff.register_user(username, tags=tags, properties=properties)

    resource_client = ff.ResourceClient(local=is_local, insecure=is_insecure)
    resource_client.apply()

    user = resource_client.get_user(username, is_local)

    assert (
        get_tags(user, is_local) == tags
        and get_properties(user, is_local) == properties
    )


@pytest.mark.parametrize(
    "is_local,is_insecure",
    [
        pytest.param(False, False, marks=pytest.mark.hosted),
        pytest.param(False, True, marks=pytest.mark.docker),
    ],
)
def test_updating_tags_and_properties_for_user(is_local, is_insecure, request):
    username = "ff_user"
    tags = ["primary_user"]
    properties = {"rotated_credentials": "yes"}

    ff.register_user(username, tags=tags, properties=properties)

    resource_client = ff.ResourceClient(local=is_local, insecure=is_insecure)
    resource_client.apply()

    ff.clear_state()

    additional_tags = ["shared_credentials"]
    additional_properties = {"credential_location": "secret_vault"}

    ff.register_user(username, tags=additional_tags, properties=additional_properties)
    resource_client.apply()

    user = resource_client.get_user(username, is_local)

    expected_tags = set(tags + additional_tags)
    expected_properties = {**properties, **additional_properties}

    actual_tags = set(get_tags(user, is_local))
    actual_properties = get_properties(user, is_local)

    assert actual_tags == expected_tags and actual_properties == expected_properties


@pytest.mark.parametrize(
    "provider_source_fxt,is_local,is_insecure",
    [
        pytest.param(
            "hosted_sql_provider_and_source", False, False, marks=pytest.mark.hosted
        ),
        pytest.param(
            "hosted_sql_provider_and_source", False, True, marks=pytest.mark.docker
        ),
    ],
)
def test_adding_tags_and_properties_to_source(
    provider_source_fxt, is_local, is_insecure, request
):
    custom_marks = [
        mark.name for mark in request.node.own_markers if mark.name != "parametrize"
    ]
    provider, source, inference_store = request.getfixturevalue(provider_source_fxt)(
        custom_marks
    )

    name = "transactions_src"
    variant = "quickstart_v2"
    tags = ["fraud_project", "primary_source"]
    properties = {"project_type": "fraud_prediction"}

    if is_local:
        provider.register_file(
            name=name,
            variant=variant,
            description="A dataset of fraudulent transactions.",
            path=f"{dir_path}/test_files/input_files/transactions.csv",
            tags=tags,
            properties=properties,
        )
    else:
        provider.register_table(
            name=name,
            table="Transactions",  # This is the table's name in Postgres
            variant=variant,
            tags=tags,
            properties=properties,
        )

    resource_client = ff.ResourceClient(local=is_local, insecure=is_insecure)
    resource_client.apply()

    source = resource_client.print_source(name, variant, is_local)

    assert (
        get_tags(source, is_local) == tags
        and get_properties(source, is_local) == properties
    )


@pytest.mark.parametrize(
    "provider_source_fxt,is_local,is_insecure",
    [
        pytest.param(
            "hosted_sql_provider_and_source", False, False, marks=pytest.mark.hosted
        ),
        pytest.param(
            "hosted_sql_provider_and_source", False, True, marks=pytest.mark.docker
        ),
    ],
)
def test_updating_tags_and_properties_for_source(
    provider_source_fxt, is_local, is_insecure, request
):
    custom_marks = [
        mark.name for mark in request.node.own_markers if mark.name != "parametrize"
    ]
    provider, source, inference_store = request.getfixturevalue(provider_source_fxt)(
        custom_marks
    )

    name = "transactions_src_2"
    variant = "quickstart_v1"
    tags = ["fraud_project", "primary_source"]
    properties = {"project_type": "fraud_prediction"}

    if is_local:
        provider.register_file(
            name=name,
            variant=variant,
            description="A dataset of fraudulent transactions.",
            path=f"{dir_path}/test_files/input_files/transactions.csv",
            tags=tags,
            properties=properties,
        )
    else:
        provider.register_table(
            name=name,
            table="Transactions",  # This is the table's name in Postgres
            variant=variant,
            tags=tags,
            properties=properties,
        )

    resource_client = ff.ResourceClient(local=is_local, insecure=is_insecure)
    resource_client.apply()

    ff.clear_state()

    additional_tags = ["v2"]
    additional_properties = {"version": "2"}

    if is_local:
        provider.register_file(
            name=name,
            variant=variant,
            description="A dataset of fraudulent transactions.",
            path=f"{dir_path}/test_files/input_files/transactions.csv",
            tags=additional_tags,
            properties=additional_properties,
        )
    else:
        provider.register_table(
            name=name,
            table="Transactions",  # This is the table's name in Postgres
            variant=variant,
            tags=additional_tags,
            properties=additional_properties,
        )

    resource_client.apply()

    source = resource_client.print_source(name, variant, is_local)

    expected_tags = set(tags + additional_tags)
    expected_properties = {**properties, **additional_properties}

    actual_tags = set(get_tags(source, is_local))
    actual_properties = get_properties(source, is_local)

    assert actual_tags == expected_tags and actual_properties == expected_properties


@pytest.mark.parametrize(
    "provider_source_fxt,is_local,is_insecure",
    [
        pytest.param(
            "hosted_sql_provider_and_source", False, False, marks=pytest.mark.hosted
        ),
        pytest.param(
            "hosted_sql_provider_and_source", False, True, marks=pytest.mark.docker
        ),
    ],
)
def test_adding_tags_and_properties_to_transformation(
    provider_source_fxt, is_local, is_insecure, request
):
    custom_marks = [
        mark.name for mark in request.node.own_markers if mark.name != "parametrize"
    ]
    provider, source, inference_store = request.getfixturevalue(provider_source_fxt)(
        custom_marks
    )
    tags = ["user_avg"]
    properties = {"aggregate_type": "avg"}
    variant = "quickstart"
    if is_local:

        @provider.df_transformation(
            variant=variant,
            inputs=[("transactions", "quickstart")],
            tags=tags,
            properties=properties,
        )
        def average_user_transaction_v1(transactions):
            return transactions.groupby("CustomerID")["TransactionAmount"].mean()

    else:

        @provider.sql_transformation(variant=variant, tags=tags, properties=properties)
        def average_user_transaction_v1():
            return "SELECT customerid as user_id, avg(transactionamount) as avg_transaction_amt from {{transactions.quickstart}} GROUP BY user_id"

    resource_client = ff.ResourceClient(local=is_local, insecure=is_insecure)
    resource_client.apply()

    source = resource_client.print_source(
        "average_user_transaction_v1", variant, is_local
    )

    assert (
        get_tags(source, is_local) == tags
        and get_properties(source, is_local) == properties
    )


@pytest.mark.parametrize(
    "provider_source_fxt,is_local,is_insecure",
    [
        pytest.param(
            "hosted_sql_provider_and_source", False, False, marks=pytest.mark.hosted
        ),
        pytest.param(
            "hosted_sql_provider_and_source", False, True, marks=pytest.mark.docker
        ),
    ],
)
def test_updating_tags_and_properties_for_transformation(
    provider_source_fxt, is_local, is_insecure, request
):
    custom_marks = [
        mark.name for mark in request.node.own_markers if mark.name != "parametrize"
    ]
    provider, source, inference_store = request.getfixturevalue(provider_source_fxt)(
        custom_marks
    )
    tags = ["user_avg"]
    properties = {"aggregate_type": "avg"}
    variant = "quickstart"
    if is_local:

        @provider.df_transformation(
            variant=variant,
            inputs=[("transactions", "quickstart")],
            tags=tags,
            properties=properties,
        )
        def average_user_transaction_v2(transactions):
            return transactions.groupby("CustomerID")["TransactionAmount"].mean()

    else:

        @provider.sql_transformation(variant=variant, tags=tags, properties=properties)
        def average_user_transaction_v2():
            return "SELECT customerid as user_id, avg(transactionamount) as avg_transaction_amt from {{transactions.quickstart}} GROUP BY user_id"

    resource_client = ff.ResourceClient(local=is_local, insecure=is_insecure)
    resource_client.apply()

    ff.clear_state()

    additional_tags = ["avg_user_transaction_v2"]
    additional_properties = {"version": "2"}

    if is_local:

        @provider.df_transformation(
            variant=variant,
            inputs=[("transactions", "quickstart")],
            tags=additional_tags,
            properties=additional_properties,
        )
        def average_user_transaction_v2(transactions):
            return transactions.groupby("CustomerID")["TransactionAmount"].mean()

    else:

        @provider.sql_transformation(
            variant=variant, tags=additional_tags, properties=additional_properties
        )
        def average_user_transaction_v2():
            return "SELECT customerid as user_id, avg(transactionamount) as avg_transaction_amt from {{transactions.quickstart}} GROUP BY user_id"

    resource_client.apply()
    source = resource_client.print_source(
        "average_user_transaction_v2", variant, is_local
    )

    expected_tags = set(tags + additional_tags)
    expected_properties = {**properties, **additional_properties}

    actual_tags = set(get_tags(source, is_local))
    actual_properties = get_properties(source, is_local)

    assert actual_tags == expected_tags and actual_properties == expected_properties


@pytest.mark.parametrize(
    "is_local,is_insecure",
    [
        pytest.param(False, False, marks=pytest.mark.hosted),
        pytest.param(False, True, marks=pytest.mark.docker),
    ],
)
def test_adding_tags_and_properties_to_entity(is_local, is_insecure, request):
    name = "cc_user"
    tags = ["customers", "user_ent"]
    properties = {
        "entity_name": "user",
        "is_user_data": "yes",
    }

    ff.register_entity(name, tags=tags, properties=properties)
    resource_client = ff.ResourceClient(local=is_local, insecure=is_insecure)
    resource_client.apply()

    entity = resource_client.get_entity(name, is_local)

    assert (
        get_tags(entity, is_local) == tags
        and get_properties(entity, is_local) == properties
    )


@pytest.mark.parametrize(
    "is_local,is_insecure",
    [
        pytest.param(False, False, marks=pytest.mark.hosted),
        pytest.param(False, True, marks=pytest.mark.docker),
    ],
)
def test_updating_tags_and_properties_for_entity(is_local, is_insecure, request):
    name = "cc_user_2"
    tags = ["customers", "user_ent"]
    properties = {
        "entity_name": "user",
        "is_user_data": "yes",
    }

    ff.register_entity(name, tags=tags, properties=properties)
    resource_client = ff.ResourceClient(local=is_local, insecure=is_insecure)
    resource_client.apply()

    ff.clear_state()

    additional_tags = ["entity_v2"]
    additional_properties = {"ent_version": "2", "updated_at": "2023-03-12"}

    ff.register_entity(name, tags=additional_tags, properties=additional_properties)
    resource_client.apply()

    entity = resource_client.get_entity(name, is_local)

    expected_tags = set(tags + additional_tags)
    expected_properties = {**properties, **additional_properties}

    actual_tags = set(get_tags(entity, is_local))
    actual_properties = get_properties(entity, is_local)

    assert actual_tags == expected_tags and actual_properties == expected_properties


@pytest.mark.parametrize(
    "provider_source_fxt,is_local,is_insecure",
    [
        pytest.param(
            "hosted_sql_provider_and_source", False, False, marks=pytest.mark.hosted
        ),
        pytest.param(
            "hosted_sql_provider_and_source", False, True, marks=pytest.mark.docker
        ),
    ],
)
def test_adding_tags_and_properties_to_feature(
    provider_source_fxt, is_local, is_insecure, request
):
    custom_marks = [
        mark.name for mark in request.node.own_markers if mark.name != "parametrize"
    ]
    provider, source, inference_store = request.getfixturevalue(provider_source_fxt)(
        custom_marks
    )
    tags = ["feat_1", "feat_2", "feat_3"]
    properties = {
        "feat_prop_key_1": "feat_prop_val_1",
        "feat_prop_key_2": "feat_prop_val_2",
    }
    # Arranges the resources context following the Quickstart pattern
    resource_client = arrange_resources(
        provider,
        source,
        inference_store,
        tags,
        properties,
        "feature",
        is_local,
        is_insecure,
    )

    feature = resource_client.print_feature("avg_transactions", "quickstart", is_local)

    assert (
        get_tags(feature, is_local) == tags
        and get_properties(feature, is_local) == properties
    )


@pytest.mark.parametrize(
    "provider_source_fxt,is_local,is_insecure",
    [
        pytest.param(
            "hosted_sql_provider_and_source", False, False, marks=pytest.mark.hosted
        ),
        pytest.param(
            "hosted_sql_provider_and_source", False, True, marks=pytest.mark.docker
        ),
    ],
)
def test_updating_tags_and_properties_for_feature(
    provider_source_fxt, is_local, is_insecure, request
):
    custom_marks = [
        mark.name for mark in request.node.own_markers if mark.name != "parametrize"
    ]
    provider, source, inference_store = request.getfixturevalue(provider_source_fxt)(
        custom_marks
    )
    tags = ["feat_1", "feat_2", "feat_3"]
    properties = {
        "feat_prop_key_1": "feat_prop_val_1",
        "feat_prop_key_2": "feat_prop_val_2",
    }
    # Arranges the resources context following the Quickstart pattern
    resource_client = arrange_resources(
        provider,
        source,
        inference_store,
        tags,
        properties,
        "feature",
        is_local,
        is_insecure,
    )

    ff.clear_state()

    additional_tags = ["feat_tag_4", "feat_tag_5"]
    additional_properties = {"feat_prop_key_3": "feat_prop_val_3"}

    resource_client = arrange_resources(
        provider,
        source,
        inference_store,
        additional_tags,
        additional_properties,
        "feature",
        is_local,
        is_insecure,
    )

    feature = resource_client.print_feature("avg_transactions", "quickstart", is_local)

    expected_tags = set(tags + additional_tags)
    expected_properties = {**properties, **additional_properties}

    actual_tags = set(get_tags(feature, is_local))
    actual_properties = get_properties(feature, is_local)

    assert actual_tags == expected_tags and actual_properties == expected_properties


@pytest.mark.parametrize(
    "provider_source_fxt,is_local,is_insecure",
    [
        pytest.param(
            "hosted_sql_provider_and_source", False, False, marks=pytest.mark.hosted
        ),
        pytest.param(
            "hosted_sql_provider_and_source", False, True, marks=pytest.mark.docker
        ),
    ],
)
def test_adding_tags_and_properties_to_label(
    provider_source_fxt, is_local, is_insecure, request
):
    custom_marks = [
        mark.name for mark in request.node.own_markers if mark.name != "parametrize"
    ]
    provider, source, inference_store = request.getfixturevalue(provider_source_fxt)(
        custom_marks
    )
    tags = ["lbl_tag_1", "lbl_tag_2", "lbl_tag_3"]
    properties = {
        "lbl_prop_key_1": "lbl_prop_val_1",
        "lbl_prop_key_1": "lbl_prop_val_2",
    }
    # Arranges the resources context following the Quickstart pattern
    resource_client = arrange_resources(
        provider,
        source,
        inference_store,
        tags,
        properties,
        "label",
        is_local,
        is_insecure,
    )

    label = resource_client.print_label("fraudulent", "quickstart", is_local)

    assert (
        get_tags(label, is_local) == tags
        and get_properties(label, is_local) == properties
    )


@pytest.mark.parametrize(
    "provider_source_fxt,is_local,is_insecure",
    [
        pytest.param(
            "hosted_sql_provider_and_source", False, False, marks=pytest.mark.hosted
        ),
        pytest.param(
            "hosted_sql_provider_and_source", False, True, marks=pytest.mark.docker
        ),
    ],
)
def test_updating_tags_and_properties_for_label(
    provider_source_fxt, is_local, is_insecure, request
):
    custom_marks = [
        mark.name for mark in request.node.own_markers if mark.name != "parametrize"
    ]
    provider, source, inference_store = request.getfixturevalue(provider_source_fxt)(
        custom_marks
    )
    tags = ["lbl_tag_1", "lbl_tag_2", "lbl_tag_3"]
    properties = {
        "lbl_prop_key_1": "lbl_prop_val_1",
        "lbl_prop_key_1": "lbl_prop_val_2",
    }
    # Arranges the resources context following the Quickstart pattern
    resource_client = arrange_resources(
        provider,
        source,
        inference_store,
        tags,
        properties,
        "label",
        is_local,
        is_insecure,
    )

    ff.clear_state()

    additional_tags = ["lbl_tag_4", "lbl_tag_5"]
    additional_properties = {"lbl_prop_key_3": "lbl_prop_val_3"}

    resource_client = arrange_resources(
        provider,
        source,
        inference_store,
        additional_tags,
        additional_properties,
        "label",
        is_local,
        is_insecure,
    )

    label = resource_client.print_label("fraudulent", "quickstart", is_local)

    expected_tags = set(tags + additional_tags)
    expected_properties = {**properties, **additional_properties}

    actual_tags = set(get_tags(label, is_local))
    actual_properties = get_properties(label, is_local)

    assert actual_tags == expected_tags and actual_properties == expected_properties


@pytest.mark.parametrize(
    "provider_source_fxt,is_local,is_insecure",
    [
        pytest.param(
            "hosted_sql_provider_and_source", False, False, marks=pytest.mark.hosted
        ),
        pytest.param(
            "hosted_sql_provider_and_source", False, True, marks=pytest.mark.docker
        ),
    ],
)
def test_adding_tags_and_properties_to_training_set(
    provider_source_fxt, is_local, is_insecure, request
):
    custom_marks = [
        mark.name for mark in request.node.own_markers if mark.name != "parametrize"
    ]
    provider, source, inference_store = request.getfixturevalue(provider_source_fxt)(
        custom_marks
    )
    tags = ["ts_tag_1", "ts_tag_2", "ts_tag_3"]
    properties = {"ts_prop_key_1": "ts_prop_val_1", "ts_prop_key_1": "ts_prop_val_2"}
    # Arranges the resources context following the Quickstart pattern
    resource_client = arrange_resources(
        provider,
        source,
        inference_store,
        tags,
        properties,
        "training-set",
        is_local,
        is_insecure,
    )

    training_set = resource_client.print_training_set(
        "fraud_training", "quickstart", is_local
    )

    assert (
        get_tags(training_set, is_local) == tags
        and get_properties(training_set, is_local) == properties
    )


@pytest.mark.parametrize(
    "provider_source_fxt,is_local,is_insecure",
    [
        pytest.param(
            "hosted_sql_provider_and_source", False, False, marks=pytest.mark.hosted
        ),
        pytest.param(
            "hosted_sql_provider_and_source", False, True, marks=pytest.mark.docker
        ),
    ],
)
def test_updating_tags_and_properties_for_training_set(
    provider_source_fxt, is_local, is_insecure, request
):
    custom_marks = [
        mark.name for mark in request.node.own_markers if mark.name != "parametrize"
    ]
    provider, source, inference_store = request.getfixturevalue(provider_source_fxt)(
        custom_marks
    )
    tags = ["ts_tag_1", "ts_tag_2", "ts_tag_3"]
    properties = {"ts_prop_key_1": "ts_prop_val_1", "ts_prop_key_1": "ts_prop_val_2"}
    # Arranges the resources context following the Quickstart pattern
    resource_client = arrange_resources(
        provider,
        source,
        inference_store,
        tags,
        properties,
        "training-set",
        is_local,
        is_insecure,
    )

    ff.clear_state()

    additional_tags = ["ts_tag_4", "ts_tag_5"]
    additional_properties = {"ts_prop_key_3": "ts_prop_val_3"}

    resource_client = arrange_resources(
        provider,
        source,
        inference_store,
        additional_tags,
        additional_properties,
        "training-set",
        is_local,
        is_insecure,
    )

    training_set = resource_client.print_training_set(
        "fraud_training", "quickstart", is_local
    )

    expected_tags = set(tags + additional_tags)
    expected_properties = {**properties, **additional_properties}

    actual_tags = set(get_tags(training_set, is_local))
    actual_properties = get_properties(training_set, is_local)

    assert actual_tags == expected_tags and actual_properties == expected_properties


@pytest.fixture(autouse=True)
def before_and_after_each(setup_teardown):
    setup_teardown()
    yield
    setup_teardown()


def arrange_resources(
    provider,
    source,
    online_store,
    tags,
    properties,
    resource_type,
    is_local,
    is_insecure,
):
    if is_local:

        @provider.df_transformation(
            variant="quickstart", inputs=[("transactions", "quickstart")]
        )
        def average_user_transaction(transactions):
            return transactions.groupby("CustomerID")["TransactionAmount"].mean()

    else:

        @provider.sql_transformation(variant="quickstart")
        def average_user_transaction():
            return "SELECT customerid as user_id, avg(transactionamount) as avg_transaction_amt from {{transactions.quickstart}} GROUP BY user_id"

    user = ff.register_entity("user")
    feature_column = "TransactionAmount" if is_local else "avg_transaction_amt"
    label_column = "IsFraud" if is_local else "isfraud"
    inference_store = provider if is_local else online_store
    feature_tags = tags if resource_type == "feature" else []
    feature_properties = properties if resource_type == "feature" else {}
    label_tags = tags if resource_type == "label" else []
    label_properties = properties if resource_type == "label" else {}

    if resource_type != "label":
        average_user_transaction.register_resources(
            entity=user,
            entity_column="CustomerID" if is_local else "user_id",
            inference_store=inference_store,
            features=[
                {
                    "name": "avg_transactions",
                    "variant": "quickstart",
                    "column": feature_column,
                    "type": "float32",
                    "tags": feature_tags,
                    "properties": feature_properties,
                },
            ],
        )

    if resource_type != "feature":
        source.register_resources(
            entity=user,
            entity_column="CustomerID" if is_local else "customerid",
            labels=[
                {
                    "name": "fraudulent",
                    "variant": "quickstart",
                    "column": label_column,
                    "type": "bool",
                    "tags": label_tags,
                    "properties": label_properties,
                },
            ],
        )

    training_set_name = "fraud_training"
    training_set_variant = "quickstart"

    if resource_type == "training-set":
        ff.register_training_set(
            training_set_name,
            training_set_variant,
            label=("fraudulent", "quickstart"),
            features=[("avg_transactions", "quickstart")],
            tags=tags,
            properties=properties,
        )

    resource_client = ff.ResourceClient(local=is_local, insecure=is_insecure)
    resource_client.apply()

    if not is_local and resource_type == "training-set":
        start = time.time()
        while True:
            time.sleep(3)
            ts = resource_client.get_training_set(
                training_set_name, training_set_variant
            )
            elapsed_wait = time.time() - start
            if (elapsed_wait >= 60) and ts.status != "READY":
                print(
                    f"Wait time for training set status exceeded; status is {ts.status}"
                )
                break
            elif ts.status == "READY":
                print(f"Training set is ready")
                break
            else:
                print(
                    f"Training set status is currently {ts.status} after {elapsed_wait} seconds ..."
                )
                continue

    return resource_client


def get_tags(resource, is_local):
    return resource["tags"] if is_local else resource.tags.tag


def get_properties(resource, is_local):
    return (
        resource["properties"]
        if is_local
        else {k: v.string_value for (k, v) in resource.properties.property.items()}
    )



from dataclasses import dataclass
from enum import Enum
from featureform.proto import metadata_pb2 as pb
from typeguard import typechecked
from os import path
from fnmatch import fnmatch


class ScalarType(Enum):
    """
    ScalarType is an enum of all the scalar types supported by Featureform.

    Attributes:
        NIL: An empty string representing no specified type.
        INT: A string representing an integer type.
        INT32: A string representing a 32-bit integer type.
        INT64: A string representing a 64-bit integer type.
        FLOAT32: A string representing a 32-bit float type.
        FLOAT64: A string representing a 64-bit float type.
        STRING: A string representing a string type.
        BOOL: A string representing a boolean type.
        DATETIME: A string representing a datetime type.
    """

    NIL = ""
    INT = "int"
    INT32 = "int32"
    INT64 = "int64"
    FLOAT32 = "float32"
    FLOAT64 = "float64"
    STRING = "string"
    BOOL = "bool"
    DATETIME = "datetime"

    @classmethod
    def has_value(cls, value):
        try:
            cls(value)
            return True
        except ValueError:
            return False

    @classmethod
    def get_values(cls):
        return [e.value for e in cls]

    def to_proto(self):
        proto_enum = self.to_proto_enum()
        return pb.ValueType(scalar=proto_enum)

    def to_proto_enum(self):
        mapping = {
            ScalarType.NIL: pb.ScalarType.NULL,
            ScalarType.INT: pb.ScalarType.INT,
            ScalarType.INT32: pb.ScalarType.INT32,
            ScalarType.INT64: pb.ScalarType.INT64,
            ScalarType.FLOAT32: pb.ScalarType.FLOAT32,
            ScalarType.FLOAT64: pb.ScalarType.FLOAT64,
            ScalarType.STRING: pb.ScalarType.STRING,
            ScalarType.BOOL: pb.ScalarType.BOOL,
            ScalarType.DATETIME: pb.ScalarType.DATETIME,
        }
        return mapping[self]

    @classmethod
    def from_proto(cls, proto_val):
        mapping = {
            pb.ScalarType.NULL: ScalarType.NIL,
            pb.ScalarType.INT: ScalarType.INT,
            pb.ScalarType.INT32: ScalarType.INT32,
            pb.ScalarType.INT64: ScalarType.INT64,
            pb.ScalarType.FLOAT32: ScalarType.FLOAT32,
            pb.ScalarType.FLOAT64: ScalarType.FLOAT64,
            pb.ScalarType.STRING: ScalarType.STRING,
            pb.ScalarType.BOOL: ScalarType.BOOL,
            pb.ScalarType.DATETIME: ScalarType.DATETIME,
        }
        return mapping[proto_val]


class ResourceStatus(str, Enum):
    """
    ResourceStatus is an enumeration representing the possible states that a
    resource may occupy within an application.

    Each status is represented as a string, which provides a human-readable
    representation for each of the stages in the lifecycle of a resource.

    Attributes:
        NO_STATUS (str): The state of a resource that cannot have another status.
        CREATED (str): The state after a resource has been successfully created.
        PENDING (str): The state indicating that the resource is in the process of being prepared, but is not yet ready.
        READY (str): The state indicating that the resource has been successfully prepared and is now ready for use.
        FAILED (str): The state indicating that an error occurred during the creation or preparation of the resource.
    """

    NO_STATUS = "NO_STATUS"
    CREATED = "CREATED"
    PENDING = "PENDING"
    READY = "READY"
    FAILED = "FAILED"

    @staticmethod
    def from_proto(proto):
        return proto.Status._enum_type.values[proto.status].name


class ComputationMode(Enum):
    PRECOMPUTED = "PRECOMPUTED"
    CLIENT_COMPUTED = "CLIENT_COMPUTED"
    STREAMING = "STREAMING"

    def __eq__(self, other: str) -> bool:
        return self.value == other

    def proto(self) -> int:
        if self == ComputationMode.PRECOMPUTED:
            return pb.ComputationMode.PRECOMPUTED
        elif self == ComputationMode.CLIENT_COMPUTED:
            return pb.ComputationMode.CLIENT_COMPUTED
        elif self == ComputationMode.STREAMING:
            return pb.ComputationMode.STREAMING
        else:
            raise ValueError(f"Unknown computation mode: {self}")


@typechecked
@dataclass
class OperationType(Enum):
    GET = 0
    CREATE = 1


@typechecked
@dataclass
class SourceType(str, Enum):
    PRIMARY_SOURCE = "PRIMARY"
    DIRECTORY = "DIRECTORY"
    DF_TRANSFORMATION = "DF"
    SQL_TRANSFORMATION = "SQL"


class FilePrefix(Enum):
    S3 = ("s3://", "s3a://")
    S3A = ("s3a://",)
    HDFS = ("hdfs://",)
    GCS = ("gs://",)
    AZURE = ("abfss://",)
    GLUE = ("glue://",)

    def __init__(self, *valid_prefixes):
        self.prefixes = valid_prefixes

    @property
    def value(self):
        return self.prefixes[0]

    def validate_file_scheme(self, file_path: str) -> (bool, str):
        if not any(file_path.startswith(prefix) for prefix in self.prefixes):
            raise Exception(
                f"File path '{file_path}' must be a full path. Must start with '{self.prefixes}'"
            )

    @staticmethod
    def validate(store_type: str, file_path: str):
        try:
            prefix = FilePrefix[store_type]
            prefix.validate_file_scheme(file_path)
        except KeyError:
            raise Exception(f"Invalid store type: {store_type}")


class FileFormat(str, Enum):
    CSV = "csv"
    PARQUET = "parquet"

    @classmethod
    def is_supported(cls, file_path: str) -> bool:
        file_name = path.basename(file_path)

        for file_format in cls:
            if fnmatch(file_name, f"*.{file_format.value}"):
                return True

        return False

    @classmethod
    def get_format(cls, file_path: str, default: str = "") -> str:
        file_name = path.basename(file_path)

        for file_format in cls:
            if fnmatch(file_name, f"*.{file_format.value}"):
                return file_format.value

        if default != "":
            return default
        raise ValueError(f"File format not supported: {file_name}")

    @classmethod
    def supported_formats(cls) -> str:
        return ", ".join([file_format.value for file_format in cls])


@typechecked
@dataclass
class ResourceType(Enum):
    # ResourceType is an enumeration representing the possible types of
    # resources that may be registered with Featureform. Each value is based
    # on OfflineResourceType in providers/offline.go

    NO_TYPE = 0
    LABEL = 1
    FEATURE = 2
    TRAINING_SET = 3
    PRIMARY = 4
    TRANSFORMATION = 5
    FEATURE_MATERIALIZATION = 6



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.
import inspect
import os
import random
import types
import warnings
from collections.abc import Iterator
from typing import List, Optional, Union

import dill
import math
import numpy as np
import pandas as pd
from featureform.proto import serving_pb2, serving_pb2_grpc
from . import GrpcClient, Model, TrainingSetVariant
from .enums import FileFormat, ResourceType
from .register import FeatureColumnResource
from .tls import insecure_channel, secure_channel

from .train_test_split import TrainTestSplit
from .version import check_up_to_date
from .grpc_client import GrpcClient


def check_feature_type(features):
    checked_features = []
    for feature in features:
        if isinstance(feature, tuple):
            checked_features.append(feature)
        elif isinstance(feature, str):
            # TODO: Need to identify how to pull the run id
            checked_features.append((feature, "default"))
        elif isinstance(feature, FeatureColumnResource):
            checked_features.append(feature.name_variant())
        elif hasattr(feature, "name_variant"):
            checked_features.append(feature.name_variant())
        else:
            raise ValueError(
                f"Invalid feature type {type(feature)}; must be a tuple, string, or FeatureColumnResource"
            )
    return checked_features


class ServingClient:
    """
    The serving client is used to retrieve training sets and features for training and serving purposes.
    **Using the Serving Client:**
    ``` py
    import featureform as ff
    from featureform import Client
    client = Client()
    # example:
    dataset = client.training_set("fraud_training", "quickstart")
    training_dataset = dataset.repeat(10).shuffle(1000).batch(8)
    for feature_batch in training_dataset:
        # Train model
    ```
    """

    def __init__(
        self, host=None, local=False, insecure=False, cert_path=None, debug=False
    ):
        # This line ensures that the warning is only raised if ServingClient is instantiated directly
        # TODO: Remove this check once ServingClient is deprecated
        is_instantiated_directed = inspect.stack()[1].function != "__init__"
        if is_instantiated_directed:
            warnings.warn(
                "ServingClient is deprecated and will be removed in future versions; use Client instead.",
                PendingDeprecationWarning,
            )
        """
        Args:
            host (str): The hostname of the Featureform instance.
            insecure (bool): True if connecting to an insecure Featureform endpoint. False if using a self-signed or public TLS certificate
            cert_path (str): The path to a public certificate if using a self-signed certificate.
        """
        if local:
            raise Exception(
                "Local mode is not supported in this version. Use featureform <= 1.12.0 for localmode"
            )

        self.impl = HostedClientImpl(host, insecure, cert_path)

    def training_set(
        self,
        name,
        variant="",
        include_label_timestamp=False,
        model: Union[str, Model] = None,
    ) -> "Dataset":
        """Return an iterator that iterates through the specified training set.

        **Examples**:
        ``` py
            client = ff.Client()
            dataset = client.training_set("fraud_training", "quickstart")
            training_dataset = dataset.repeat(10).shuffle(1000).batch(8)
            for feature_batch in training_dataset:
                # Train model
        ```

        Args:
            name (str): Name of training set to be retrieved
            variant (str): Variant of training set to be retrieved

        Returns:
            training_set (Dataset): A training set iterator
        """
        if isinstance(name, TrainingSetVariant):
            variant = name.variant
            name = name.name
        return self.impl.training_set(name, variant, include_label_timestamp, model)

    def features(
        self, features, entities, model: Union[str, Model] = None, params: list = None
    ):
        """Returns the feature values for the specified entities.

        **Examples**:
        ``` py
            client = ff.Client()
            fpf = client.features([("avg_transactions", "quickstart")], {"user": "C1410926"})
            # Run features through model
        ```

        Args:
            features (list[(str, str)], list[str]): List of Name Variant Tuples
            entities (dict): Dictionary of entity name/value pairs

        Returns:
            features (numpy.Array): An Numpy array of feature values in the order given by the inputs
        """
        features = check_feature_type(features)
        return self.impl.features(features, entities, model, params)

    def close(self):
        """Closes the connection to the Featureform instance."""
        self.impl.close()

    def batch_features(self, features):
        """
        Return an iterator that iterates over each entity and corresponding features in feats.
        **Example:**
        ```py title="definitions.py"
        for feature_values in client.batch_features([("feature1", "variant"), ("feature2", "variant"), ("feature3", "variant")]):
            print(feature_values)
        ```

        Args:
            features (List[NameVariant]): The features to iterate over

        Returns:
            iterator: An iterator of entity and feature values

        """
        if len(features) == 0:
            raise ValueError("No features provided")
        feature_tuples = check_feature_type(features)
        return self.impl.batch_features(feature_tuples)


class HostedClientImpl:
    def __init__(self, host=None, insecure=False, cert_path=None, debug=False):
        host = host or os.getenv("FEATUREFORM_HOST")
        if host is None:
            raise ValueError(
                "The `host` parameter must be passed or the environment"
                " variable FEATUREFORM_HOST must be set."
            )
        check_up_to_date(False, "serving")
        self._channel = self._create_channel(host, insecure, cert_path)
        self._stub = GrpcClient(
            serving_pb2_grpc.FeatureStub(self._channel),
            debug=debug,
            insecure=insecure,
            host=host,
        )

    @staticmethod
    def _create_channel(host, insecure, cert_path):
        if insecure:
            return insecure_channel(host)
        else:
            return secure_channel(host, cert_path)

    def training_set(
        self, name, variation, include_label_timestamp, model: Union[str, Model] = None
    ):
        training_set_stream = TrainingSetStream(self._stub, name, variation, model)
        return Dataset(training_set_stream)

    def features(
        self, features, entities, model: Union[str, Model] = None, params: list = None
    ):
        req = serving_pb2.FeatureServeRequest()
        for name, values in entities.items():
            entity_proto = req.entities.add()
            entity_proto.name = name
            if isinstance(values, list):
                for value in values:  # Assuming 'values' is a list of strings
                    entity_proto.values.append(value)
            elif isinstance(values, str):
                entity_proto.values.append(values)
            else:
                raise ValueError(
                    "Entity values must be either a string or a list of strings"
                )
        for name, variation in features:
            feature_id = req.features.add()
            feature_id.name = name
            feature_id.version = variation
        if model is not None:
            req.model.name = model if isinstance(model, str) else model.name
        resp = self._stub.FeatureServe(req)

        deprecated_values = resp.values
        value_lists = (
            deprecated_values if len(deprecated_values) > 0 else resp.value_lists
        )

        feature_values = []
        for val_list in value_lists:
            entity_values = []
            for val in val_list.values:
                parsed_value = parse_proto_value(val)
                value_type = type(parsed_value)

                # Ondemand features are returned as a byte array
                # which holds the pickled function
                if value_type == bytes:
                    code = dill.loads(bytearray(parsed_value))
                    func = types.FunctionType(code, globals(), "transformation")
                    parsed_value = func(self, params, entities)
                # Vector features are returned as a Vector32 proto due
                # to the inability to use the `repeated` keyword in
                # in a `oneof` field
                elif value_type == serving_pb2.Vector32:
                    parsed_value = parsed_value.value
                entity_values.append(parsed_value)

            # If theres only one entity row, only return that row
            if len(value_lists) == 1:
                return entity_values
            feature_values.append(entity_values)

        # if only one entity is requested, return a flat list
        if len(entities.keys()) == 1 and type(feature_values[0]) is list:
            return [j for sub in feature_values for j in sub]

        return feature_values

    def batch_features(self, features):
        return FeatureSetIterator(self._stub, features)

    def _get_source_as_df(self, name, variant, limit):
        columns = self._get_source_columns(name, variant)
        data = self._get_source_data(name, variant, limit)
        return pd.DataFrame(data=data, columns=columns)

    def _get_source_data(self, name, variant, limit):
        id = serving_pb2.SourceID(name=name, version=variant)
        req = serving_pb2.SourceDataRequest(id=id, limit=limit)
        resp = self._stub.SourceData(req)
        data = []
        for rows in resp:
            row = [getattr(r, r.WhichOneof("value")) for r in rows.rows]
            data.append(row)
        return data

    def _get_source_columns(self, name, variant):
        id = serving_pb2.SourceID(name=name, version=variant)
        req = serving_pb2.SourceDataRequest(id=id)
        resp = self._stub.SourceColumns(req)
        # The Python type of resp.columns is <class 'google._upb._message.RepeatedScalarContainer'>
        # which is not a "recognized" type by pandas internal type check, which resulted in the following error:
        # `Index(...) must be called with a collection of some kind`
        # To avoid this issue, we convert the resp.columns to a Python list
        return list(resp.columns)

    def nearest(self, name, variant, vector, k):
        id = serving_pb2.FeatureID(name=name, version=variant)
        vec = serving_pb2.Vector32(value=vector)
        req = serving_pb2.NearestRequest(id=id, vector=vec, k=k)
        resp = self._stub.Nearest(req)
        return resp.entities

    def location(self, name: str, variant: str, resource_type: ResourceType) -> str:
        req = serving_pb2.ResourceIdRequest()
        req.name = name
        req.variant = variant
        req.type = resource_type.value

        resp = self._stub.GetResourceLocation(req)

        return resp.location

    def close(self):
        self._channel.close()


class TrainingSetStream(Iterator):
    def __init__(self, stub, name, version, model: Union[str, Model] = None):
        req = serving_pb2.TrainingDataRequest()
        req.id.name = name
        req.id.version = version
        if model is not None:
            req.model.name = model if isinstance(model, str) else model.name
        self.name = name
        self.version = version
        self._stub = stub
        self._req = req
        self._iter = stub.TrainingData(req)

    def __iter__(self):
        return self

    def __next__(self):
        return Row(next(self._iter))

    def restart(self):
        self._iter = self._stub.TrainingData(self._req)


class LocalStream:
    def __init__(self, datalist, include_label_timestamp):
        self._datalist = datalist
        self._iter = iter(datalist)
        self._include_label_timestamp = include_label_timestamp

    def __iter__(self):
        return iter(self._iter)

    def __next__(self):
        return LocalRow(next(self._iter), self._include_label_timestamp)

    def restart(self):
        self._iter = iter(self._datalist)


class Repeat:
    def __init__(self, repeat_num, stream):
        self.repeat_num = repeat_num
        self._stream = stream

    def __iter__(self):
        return self

    def __next__(self):
        try:
            next_val = next(self._stream)
        except StopIteration:
            self.repeat_num -= 1
            if self.repeat_num >= 0:
                self._stream.restart()
                next_val = next(self._stream)
            else:
                raise

        return next_val


class Shuffle:
    def __init__(self, buffer_size, stream):
        self.buffer_size = buffer_size
        self._shuffled_data_list = []
        self._stream = stream
        self.__setup_buffer()

    def __setup_buffer(self):
        try:
            for _ in range(self.buffer_size):
                self._shuffled_data_list.append(next(self._stream))
        except StopIteration:
            pass

    def restart(self):
        self._stream.restart()
        self.__setup_buffer()

    def __iter__(self):
        return self

    def __next__(self):
        if len(self._shuffled_data_list) == 0:
            raise StopIteration
        random_index = random.randrange(len(self._shuffled_data_list))
        next_row = self._shuffled_data_list.pop(random_index)

        try:
            self._shuffled_data_list.append(next(self._stream))
        except StopIteration:
            pass

        return next_row


class Batch:
    def __init__(self, batch_size, stream):
        self.batch_size = batch_size
        self._stream = stream

    def restart(self):
        self._stream.restart()

    def __iter__(self):
        return self

    def __next__(self):
        rows = BatchRow()
        for _ in range(self.batch_size):
            try:
                next_row = next(self._stream)
                rows.append(next_row)
            except StopIteration:
                if len(rows) == 0:
                    raise
                return rows
        return rows


class Dataset:
    def __init__(self, stream, dataframe=None):
        """Repeats the Dataset for the specified number of times

        Args:
            stream (Iterator): An iterable object.

        Returns:
            self (Dataset): Returns a Dataset created from the iterable object.
        """
        self._stream = stream
        self._dataframe = dataframe

    def train_test_split(
        self,
        test_size: float = 0,
        train_size: float = 0,
        shuffle: bool = True,
        random_state: Optional[int] = None,
        batch_size: int = 1,
    ):
        """
        (This functionality is currently only available for Clickhouse).

        Splits an existing training set into training and testing iterators. The split is processed on the underlying
        provider and calculated at serving time.

        **Examples**:

        ``` py
        import featureform as ff
        client = ff.Client()
        train, test = client
            .training_set("fraud_training", "v1")
            .train_test_split(
                test_size=0.7,
                train_size=0.3,
                shuffle=True,
                random_state=None,
                batch_size=5
            )

        for features, label in train:
            print(features)
            print(label)
            clf.partial_fit(features, label)

        for features, label in test:
            print(features)
            print(label)
            clf.score(features, label)


        # TRAIN OUTPUT
        # np.array([
        #   [1, 1, 3],
        #   [5, 1, 2],
        #   [7, 6, 5],
        #   [8, 3, 3],
        #   [5, 2, 2],
        # ])
        # np.array([2, 4, 2, 3, 4])
        # np.array([
        #   [3, 1, 2],
        #   [5, 4, 5],
        # ])
        # np.array([6, 7])

        # TEST OUTPUT
        # np.array([
        #   [5, 1, 3],
        #   [4, 3, 1],
        #   [6, 6, 7],
        # ])
        # np.array([4, 6, 7])

        ```

        Args:
            test_size (float): The ratio of test set size to train set size. Must be a value between 0 and 1. If excluded
                it will be the complement to the train_size. One of test_size or train_size must be specified.
            train_size (float): The ratio of train set size to train set size. Must be a value between 0 and 1. If excluded
                it will be the complement to the test_size. One of test_size or train_size must be specified.
            shuffle (bool): Whether to shuffle the dataset before splitting.
            random_state (Optional[int]): A random state to shuffle the dataset. If None, the dataset will be shuffled
                randomly on every call. If >0, the value will be used a seed to create random shuffle that can be repeated
                if subsequent calls use the same seed.
            batch_size (int): The size of the batch to return from the iterator. Must be greater than 0.

        Returns:
            train (Iterator): An iterator for training values.
            test (Iterator): An iterator for testing values.
        """
        if batch_size < 1:
            raise ValueError("batch_size must be 1 or greater")

        if random_state is not None and random_state < 0:
            raise ValueError("random_state must be 0 or greater")

        test_size, train_size = self.validate_test_size(test_size, train_size)

        name = self._stream.name
        variant = self._stream.version
        stub = self._stream._stub
        model = self._stream.model if hasattr(self._stream, "model") else None

        train, test = TrainTestSplit(
            stub=stub,
            name=name,
            version=variant,
            model=model,
            test_size=test_size,
            train_size=train_size,
            shuffle=shuffle,
            random_state=(0 if random_state is None else random_state),
            batch_size=batch_size,
        ).split()

        return train, test

    @staticmethod
    def validate_test_size(test_size, train_size):
        # Validate ranges
        if not 0 <= test_size <= 1:
            raise ValueError("test_size must be between 0 and 1")
        if not 0 <= train_size <= 1:
            raise ValueError("train_size must be between 0 and 1")

        # If both are specified but don't sum to 1, it's an error
        if test_size != 0 and train_size != 0 and test_size + train_size != 1:
            raise ValueError("test_size + train_size must equal 1 if both are non-zero")

        # Auto-adjust if one is 0
        if test_size == 0 and train_size > 0:
            test_size = 1 - train_size
        elif train_size == 0 and test_size > 0:
            train_size = 1 - test_size

        return test_size, train_size

    def dataframe(self, spark_session=None):
        """Returns the training set as a Pandas DataFrame or Spark DataFrame.
        Args:
        - spark_session: Optional(SparkSession)


        **Examples**:
        ``` py
            client = Client()
            df = client.training_set("fraud_training", "v1").dataframe()
            print(df.head())
            # Output:
            #                feature_avg_transactions_default             label
            # 0                                25.0                       False
            # 1                             27999.0                       False
            # 2                               459.0                       False
            # 3                              2060.0                        True
            # 4                              1762.5                       False
        ```

        Returns:
            df: Union[pd.DataFrame, pyspark.sql.DataFrame] A DataFrame containing the training set.
        """

        if self._dataframe is not None:
            return self._dataframe
        elif spark_session is not None:
            req = serving_pb2.ResourceIdRequest()
            req.name = self._stream.name
            req.variant = self._stream.version
            req.type = ResourceType.TRAINING_DATA.value
            resp = self._stream._stub.ResourceLocation(req)

            file_format = FileFormat.get_format(resp.location, default="parquet")
            location = self._sanitize_location(resp.location)
            self._dataframe = self._get_spark_dataframe(
                spark_session, file_format, location
            )
            return self._dataframe
        else:
            name = self._stream.name
            variant = self._stream.version
            stub = self._stream._stub
            id = serving_pb2.TrainingDataID(name=name, version=variant)
            req = serving_pb2.TrainingDataRequest(id=id)
            cols = stub.TrainingDataColumns(req)
            data = [r.to_dict(cols.features, cols.label) for r in self._stream]
            self._dataframe = pd.DataFrame(
                data=data, columns=[*cols.features, cols.label]
            )
            self._dataframe.rename(columns={cols.label: "label"}, inplace=True)
            return self._dataframe

    def _sanitize_location(self, location: str) -> str:
        # Returns the location directory rather than a single file if it is part-file.
        # Also converts s3:// to s3a:// if necessary.

        # If the location is a part-file, we want to get the directory instead.
        is_individual_part_file = location.split("/")[-1].startswith("part-")
        if is_individual_part_file:
            location = "/".join(location.split("/")[:-1])

        # If the schema is s3://, we want to convert it to s3a://.
        location = (
            location.replace("s3://", "s3a://")
            if location.startswith("s3://")
            else location
        )

        return location

    def _get_spark_dataframe(self, spark, file_format, location):
        if file_format not in FileFormat.supported_formats():
            raise Exception(
                f"file type '{file_format}' is not supported. Please use 'csv' or 'parquet'"
            )

        try:
            df = (
                spark.read.option("header", "true")
                .option("recursiveFileLookup", "true")
                .format(file_format)
                .load(location)
            )

            label_column_name = ""
            for col in df.columns:
                if col.startswith("Label__"):
                    label_column_name = col
                    break

            if label_column_name != "":
                df = df.withColumnRenamed(label_column_name, "Label")

        except Exception as e:
            raise Exception(
                f"please make sure the spark session has ability to read '{location}': {e}"
            )

        return df

    def from_dataframe(dataframe, include_label_timestamp):
        stream = LocalStream(dataframe.values.tolist(), include_label_timestamp)
        return Dataset(stream, dataframe)

    def pandas(self):
        warnings.warn(
            "Dataset.pandas() is deprecated and will be removed in future versions; use Dataset.dataframe() instead.",
            PendingDeprecationWarning,
        )
        if self._dataframe is None:
            _ = self.dataframe()
        return self._dataframe

    def repeat(self, num):
        """Repeats the Dataset for the specified number of times

        **Examples**:
        ``` py
            client = ff.Client()
            dataset = client.training_set("fraud_training", "quickstart")
            training_dataset = dataset.repeat(10) # Repeats data 10 times
            for feature_batch in training_dataset:
                # Train model
        ```

        Args:
            num (int): The number of times the dataset will be repeated

        Returns:
            self (Dataset): Returns the current Dataset
        """
        if num <= 0:
            raise Exception("Must repeat 1 or more times")
        self._stream = Repeat(num, self._stream)
        if self._dataframe is not None:
            temp_df = self._dataframe
            for i in range(num):
                self._dataframe = pd.concat([self._dataframe, temp_df])
        return self

    def shuffle(self, buffer_size):
        """Swaps random rows within the Dataset.

        **Examples**:
        ``` py
            client = ff.Client()
            dataset = client.training_set("fraud_training", "quickstart")
            training_dataset = dataset.shuffle(100) # Swaps 100 Rows
            for feature_batch in training_dataset:
                # Train model
        ```

        Args:
            buffer_size (int): The number of Dataset rows to be randomly swapped

        Returns:
            self (Dataset): Returns the current Dataset
        """
        if buffer_size <= 0:
            raise Exception("Buffer size must be greater than or equal to 1")
        self._stream = Shuffle(buffer_size, self._stream)
        if self._dataframe is not None:
            self._dataframe = self._dataframe.sample(frac=1)
        return self

    def batch(self, batch_size):
        """Creates a batch row in the Dataset.

        **Examples**:
        ``` py
            client = ff.Client()
            dataset = client.training_set("fraud_training", "quickstart")
            training_dataset = dataset.batch(8) # Creates a batch of 8 Datasets for each row
            for feature_batch in training_dataset:
                # Train model
        ```

        Args:
            batch_size (int): The number of items to be added to each batch

        Returns:
            self (Dataset): Returns the current Dataset
        """
        if batch_size <= 0:
            raise Exception("Batch size must be greater than or equal to 1")
        self._stream = Batch(batch_size, self._stream)
        if self._dataframe is not None:
            self._dataframe = np.array_split(
                self._dataframe, math.ceil(len(self._dataframe) // batch_size)
            )
        return self

    def __iter__(self):
        return self

    def __next__(self):
        next_val = next(self._stream)
        return next_val


class Row:
    def __init__(self, proto_row):
        self._types = [proto_type_to_np_type(feature) for feature in proto_row.features]
        self._features = np.array(
            [parse_proto_value(feature) for feature in proto_row.features],
            dtype=get_numpy_array_type(self._types),
        )
        self._label = parse_proto_value(proto_row.label)
        self._row = np.append(self._features, self._label)

    def features(self):
        return np.array([self._row[:-1]])

    def label(self):
        return np.array([self._label])

    def to_numpy(self):
        return self._row

    def to_dict(self, feature_columns: List[str], label_column: str):
        row_dict = dict(zip(feature_columns, self._features))
        row_dict[label_column] = self._label
        return row_dict

    def __repr__(self):
        return "Features: {} , Label: {}".format(self.features(), self.label())


class LocalRow:
    def __init__(self, row_list, include_label_timestamp):
        """
        If include_label_timestamp is true then we want the label to equal to the
        last two columns in the list. Otherwise, just the label will be the last column only.
        """

        self._features = row_list[:-2] if include_label_timestamp else row_list[:-1]
        self._row = row_list
        self._label = row_list[-2:] if include_label_timestamp else row_list[-1]

    def features(self):
        return [self._features]

    def label(self):
        return [self._label]

    def to_numpy(self):
        return np.array(self._row)

    def __repr__(self):
        return "Features: {} , Label: {}".format(self.features(), self.label())


class BatchRow:
    def __init__(self, rows=None):
        self._features = []
        self._labels = []
        if rows is None:
            rows = []
        self._rows = rows
        for row in rows:
            self.append(row)

    def append(self, row):
        self._features.append(row.features()[0])
        self._labels.append(row.label()[0])
        self._rows.append(row)

    def features(self):
        return self._features

    def label(self):
        return self._labels

    def to_list(self):
        return self._rows

    def __len__(self):
        return len(self._rows)


def parse_proto_value(value):
    """parse_proto_value is used to parse the one of Value message"""
    return getattr(value, value.WhichOneof("value"))


def proto_type_to_np_type(value):
    type_mapping = {
        "str_value": str,
        "int_value": np.int32,
        "int32_value": np.int32,
        "int64_value": np.int64,
        "float_value": np.float32,
        "double_value": np.float64,
        "bool_value": bool,
    }
    return type_mapping[value.WhichOneof("value")]


def get_numpy_array_type(types):
    if all(i == types[0] for i in types):
        return types[0]
    else:
        return "O"


class FeatureSetIterator:
    def __init__(self, stub, features):
        req = serving_pb2.BatchFeatureServeRequest()
        for name, variant in features:
            feature_id = req.features.add()
            feature_id.name = name
            feature_id.version = variant
        self._stub = stub
        self._req = req
        self._iter = stub.BatchFeatureServe(req)

    def __iter__(self):
        return self

    def __next__(self):
        return FeatureSetRow(next(self._iter)).to_tuple()

    def restart(self):
        self._iter = self._stub.BatchFeatureServe(self._req)


class FeatureSetRow:
    def __init__(self, proto_row):
        self._features = [parse_proto_value(feature) for feature in proto_row.features]
        self._entity = parse_proto_value(proto_row.entity)
        self._row = [self._entity, self._features]

    def features(self):
        return self._row[1]

    def entity(self):
        return [self._entity]

    def to_numpy(self):
        return np.array(self._row)

    def to_tuple(self):
        return tuple((self._entity, self._features))

    def to_dict(self, feature_columns: List[str], entity_column: str):
        row_dict = dict(zip(feature_columns, self._features))
        row_dict[entity_column] = self._entity
        return row_dict

    def __repr__(self):
        return "Features: {} , Entity: {}".format(self.features(), self.entity())



import numpy
from .enums import ScalarType
from typing import Union

pd_to_ff_datatype = {
    numpy.dtype("float64"): ScalarType.FLOAT64,
    numpy.dtype("float32"): ScalarType.FLOAT32,
    numpy.dtype("int64"): ScalarType.INT64,
    numpy.dtype("int"): ScalarType.INT,
    numpy.dtype("int32"): ScalarType.INT32,
    numpy.dtype("O"): ScalarType.STRING,
    numpy.dtype("str"): ScalarType.STRING,
    numpy.dtype("bool"): ScalarType.BOOL,
}


class VectorType:
    def __init__(
        self, scalarType: Union[ScalarType, str], dims: int, is_embedding: bool = False
    ):
        self.scalarType = (
            scalarType if isinstance(scalarType, ScalarType) else ScalarType(scalarType)
        )
        self.dims = dims
        self.is_embedding = is_embedding

    def to_proto(self):
        proto_vec = pb.VectorType(
            scalar=self.scalarType.to_proto_enum(),
            dimension=self.dims,
            is_embedding=self.is_embedding,
        )
        return pb.ValueType(vector=proto_vec)

    @classmethod
    def from_proto(cls, proto_val):
        proto_vec = protoVal.vector_type
        scalar = ScalarType.from_proto(proto_vec.scalar)
        dims = proto_vec.dimension
        is_embedding = proto_vec.is_embedding
        return VectorType(scalar, dims, is_embedding)


def type_from_proto(proto_val):
    value_type = proto_val.WhichOneof("Type")
    if value_type == "scalar":
        return ScalarType.from_proto(proto_val.scalar)
    elif value_type == "vector":
        return VectorType.from_proto(proto_val.vector)
    else:
        return ScalarType.NIL



two_row_spacing = "{:<30} {:<25}"
three_row_spacing = "{:<30} {:<30} {:<30}"
four_row_spacing = "{:<30} {:<30} {:<30} {:<30}"
five_row_spacing = "{:<30} {:<30} {:<30} {:<30} {:<30}"
divider = "-----------------------------------------------"


def format_rows(
    format_obj,
    format_obj_2=None,
    format_obj_3=None,
    format_obj_4=None,
    format_obj_5=None,
):
    # Base case for when `format_obj` is a string
    if format_obj_2 is None and type(format_obj) == str:
        print(format_obj)
    elif format_obj_2 is None:
        for s in format_obj:
            format_rows(*s)
    elif format_obj_2 is not None and format_obj_3 is None:
        print(two_row_spacing.format(format_obj, format_obj_2))
    elif format_obj_2 is not None and format_obj_3 is not None and format_obj_4 is None:
        print(three_row_spacing.format(format_obj, format_obj_2, format_obj_3))
    elif (
        format_obj_2 is not None
        and format_obj_3 is not None
        and format_obj_4 is not None
        and format_obj_5 is None
    ):
        print(
            four_row_spacing.format(
                format_obj, format_obj_2, format_obj_3, format_obj_4
            )
        )
    else:
        print(
            five_row_spacing.format(
                format_obj, format_obj_2, format_obj_3, format_obj_4, format_obj_5
            )
        )


def format_pg(s=""):
    print(divider)
    print(s)


def progress_bar(total, current, prefix="", suffix="", length=30, fill="#"):
    import sys

    if total == 0:
        return

    percent = current / total
    filled_length = int(length * percent)
    bar = fill * filled_length + "-" * (length - filled_length)
    sys.stdout.write("\r{} |{}| {}% {}".format(prefix, bar, int(percent * 100), suffix))
    sys.stdout.flush()



import sys

sys.path.insert(0, "client/src/")
import featureform as ff
from featureform.resources import (
    PostgresConfig,
    RedshiftConfig,
    ClickHouseConfig,
    GCPCredentials,
    BigQueryConfig,
    CassandraConfig,
    AWSStaticCredentials,
    DynamodbConfig,
    RedisConfig,
    MongoDBConfig,
    FirestoreConfig,
    SnowflakeConfig,
    PineconeConfig,
    S3StoreConfig,
    EMRCredentials,
    SparkConfig,
    AzureFileStoreConfig,
    GCSFileStoreConfig,
    HDFSConfig,
    BasicCredentials,
    OnlineBlobConfig,
    DatabricksCredentials,
    SparkCredentials,
)
from featureform.proto import metadata_pb2 as pb

#  Offline Providers
postgres_fields = {
    "name": "my_postgres",
    "description": "some description",
    "type": "POSTGRES_OFFLINE",
    "software": "provider",
    "team": "some team",
    "tags": ["tag1", "tag2"],
    "properties": {},
    "host": "my_host",
    "port": "1234",
    "database": "my_database",
    "user": "me",
    "password": "mypassword",
    "sslmode": "disabled",
}

clickhouse_fields = {
    "name": "my_clickhouse",
    "description": "some description",
    "type": "CLICKHOUSE_OFFLINE",
    "software": "provider",
    "team": "some team",
    "tags": ["tag1", "tag2"],
    "properties": {},
    "host": "my_host",
    "port": 1234,
    "database": "my_database",
    "user": "me",
    "password": "mypassword",
    "sslmode": True,
}

redshift_fields = {
    "name": "my_redshift",
    "description": "some description",
    "type": "REDSHIFT_OFFLINE",
    "software": "provider",
    "team": "some team",
    "tags": ["tag1", "tag2"],
    "properties": {},
    "host": "my_host",
    "port": "1234",
    "database": "my_database",
    "user": "me",
    "password": "mypassword",
    "sslmode": "disabled",
}

bigquery_fields = {
    "name": "my_bigquery",
    "project_id": "my_project",
    "type": "BIGQUERY_OFFLINE",
    "dataset_id": "my_dataset",
    "credentials_path": "client/tests/test_files/bigquery_dummy_credentials.json",
    "description": "some description",
    "team": "some team",
    "tags": ["tag1", "tag2"],
    "properties": {},
}

snowflake_legacy_fields = {
    "name": "my_snowflake_legacy",
    "type": "SNOWFLAKE_LEGACY_OFFLINE",
    "username": "me",
    "password": "mypassword",
    "account_locator": "my_account_locator",
    "database": "my_database",
    "schema": "PUBLIC",
    "description": "some description",
    "team": "some team",
    "warehouse": "my_warehouse",
    "role": "my_role",
    "tags": ["tag1", "tag2"],
    "properties": {},
}

snowflake_fields = {
    "name": "my_snowflake",
    "type": "SNOWFLAKE_OFFLINE",
    "username": "me",
    "password": "mypassword",
    "account": "my_account",
    "organization": "my_organization",
    "database": "my_database",
    "schema": "PUBLIC",
    "description": "some description",
    "team": "some team",
    "warehouse": "my_warehouse",
    "role": "my_role",
    "tags": ["tag1", "tag2"],
    "properties": {},
}

# Online Providers
cassandra_fields = {
    "name": "my_cassandra",
    "type": "CASSANDRA_ONLINE",
    "host": "my_host",
    "port": 1234,
    "username": "me",
    "password": "mypassword",
    "keyspace": "my_keyspace",
    "consistency": "TWO",
    "replication": 4,
    "description": "some description",
    "team": "some team",
    "tags": ["tag1", "tag2"],
    "properties": {},
}

firestore_fields = {
    "name": "my_firestore",
    "type": "FIRESTORE_ONLINE",
    "collection": "my_collection",
    "project_id": "my_project",
    "credentials_path": "client/tests/test_files/bigquery_dummy_credentials.json",
    "description": "some description",
    "team": "some team",
    "tags": ["tag1", "tag2"],
    "properties": {},
}

dynamodb_fields = {
    "name": "my_dynamodb",
    "type": "DYNAMODB_ONLINE",
    "aws_access_key_id": "my_access_key",
    "aws_secret_access_key": "my_secret",
    "region": "us-west-2",
    "should_import_from_s3": False,
    "description": "some description",
    "team": "some team",
    "tags": ["tag1", "tag2"],
    "properties": {},
}

redis_fields = {
    "name": "my_redis",
    "host": "my_redis_host",
    "port": 1234,
    "password": "my_password",
    "database": 0,
    "type": "REDIS_ONLINE",
    "description": "some description",
    "team": "some team",
    "tags": ["tag1", "tag2"],
    "properties": {},
}

mongodb_fields = {
    "name": "my_mongodb",
    "type": "MONGODB_ONLINE",
    "username": "me",
    "password": "mypassword",
    "database": "my_database",
    "host": "my_host",
    "port": "1234",
    "throughput": 1321,
    "description": "some description",
    "team": "some team",
    "tags": ["tag1", "tag2"],
    "properties": {},
}

pinecone_fields = {
    "name": "my_pinecone",
    "type": "PINECONE_ONLINE",
    "project_id": "my_project",
    "api_key": "my_api_key",
    "environment": "my_environment",
    "description": "some description",
    "team": "some team",
    "tags": ["tag1", "tag2"],
    "properties": {},
}

#  Filestores
s3_fields = {
    "name": "my_s3",
    "access_key": "my_access_key",
    "secret_key": "my_secret",
    "bucket_region": "us-west-2",
    "bucket_name": "my_bucket",
    "path": "my_path",
    "description": "some description",
    "team": "some team",
    "tags": ["tag1", "tag2"],
    "properties": {},
}

azure_fields = {
    "name": "my_azure",
    "account_name": "my_account",
    "account_key": "my_key",
    "container_name": "my_container",
    "root_path": "my_path",
    "description": "some description",
    "team": "some team",
    "tags": ["tag1", "tag2"],
    "properties": {},
}

gcs_fields = {
    "name": "my_gcs",
    "bucket_name": "my_bucket",
    "bucket_path": "my_path",
    "project_id": "my_project",
    "credentials_path": "client/tests/test_files/bigquery_dummy_credentials.json",
    "description": "some description",
    "team": "some team",
    "tags": ["tag1", "tag2"],
    "properties": {},
}

hdfs_fields = {
    "name": "my_hdfs",
    "host": "my_host",
    "port": "1234",
    "path": "my_path",
    "hdfs_site_file": "client/tests/test_files/yarn_files/yarn-site.xml",
    "core_site_file": "client/tests/test_files/yarn_files/core-site.xml",
    "username": "me",
    "password": "mypassword",
    "description": "some description",
    "team": "some team",
    "tags": ["tag1", "tag2"],
    "properties": {},
}

# Executors
emr_fields = {
    "cluster_id": "my_cluster_id",
    "cluster_region": "us-west-2",
    "access_key": "my_access_key",
    "secret_key": "my_secret",
}

databricks_fields = {
    "username": "my_username",
    "password": "my_password",
    "cluster_id": "1003-211215-n0pjeh9x",
}

generic_fields = {
    "master": "local",
    "deploy_mode": "client",
    "python_version": "3.10",
    "core_site_path": "client/tests/test_files/yarn_files/core-site.xml",
    "yarn_site_path": "client/tests/test_files/yarn_files/yarn-site.xml",
}

# Spark
spark_fields = {
    "name_emr_s3": "my_spark_emr_s3",
    "name_emr_hdfs": "my_spark_emr_hdfs",
    "name_emr_gcs": "my_spark_emr_gcs",
    "name_emr_azure": "my_spark_emr_azure",
    "name_databricks_s3": "my_spark_databricks_s3",
    "name_databricks_hdfs": "my_spark_databricks_hdfs",
    "name_databricks_gcs": "my_spark_databricks_gcs",
    "name_databricks_azure": "my_spark_databricks_azure",
    "name_generic_s3": "my_spark_generic_s3",
    "name_generic_hdfs": "my_spark_generic_hdfs",
    "name_generic_gcs": "my_spark_generic_gcs",
    "name_generic_azure": "my_spark_generic_azure",
    "description": "some description",
    "team": "some team",
    "tags": ["tag1", "tag2"],
    "properties": {},
}

emr_executor = EMRCredentials(
    emr_cluster_id=emr_fields["cluster_id"],
    emr_cluster_region=emr_fields["cluster_region"],
    credentials=AWSStaticCredentials(
        access_key=emr_fields["access_key"], secret_key=emr_fields["secret_key"]
    ),
)

databricks_executor = DatabricksCredentials(
    username=databricks_fields["username"],
    password=databricks_fields["password"],
    # host=databricks_fields["host"],
    # token=databricks_fields["token"],
    cluster_id=databricks_fields["cluster_id"],
)

generic_executor = SparkCredentials(
    master=generic_fields["master"],
    deploy_mode=generic_fields["deploy_mode"],
    python_version=generic_fields["python_version"],
    core_site_path=generic_fields["core_site_path"],
    yarn_site_path=generic_fields["yarn_site_path"],
)

s3_store = S3StoreConfig(
    bucket_path=s3_fields["bucket_name"],
    bucket_region=s3_fields["bucket_region"],
    credentials=AWSStaticCredentials(
        access_key=s3_fields["access_key"], secret_key=s3_fields["secret_key"]
    ),
    path=s3_fields["path"],
)

hdfs_store = HDFSConfig(
    host=hdfs_fields["host"],
    port=hdfs_fields["port"],
    path=hdfs_fields["path"],
    hdfs_site_file=hdfs_fields["hdfs_site_file"],
    core_site_file=hdfs_fields["core_site_file"],
    credentials=BasicCredentials(
        username=hdfs_fields["username"], password=hdfs_fields["password"]
    ),
)

gcs_store = GCSFileStoreConfig(
    bucket_name=gcs_fields["bucket_name"],
    bucket_path=gcs_fields["bucket_path"],
    credentials=GCPCredentials(
        project_id=gcs_fields["project_id"],
        credentials_path=gcs_fields["credentials_path"],
    ),
)

azure_store = AzureFileStoreConfig(
    account_name=azure_fields["account_name"],
    account_key=azure_fields["account_key"],
    container_name=azure_fields["container_name"],
    root_path=azure_fields["root_path"],
)
blob_store = OnlineBlobConfig(store_type="AZURE", store_config=azure_store.config())


def postgres_proto():
    return pb.Provider(
        name=postgres_fields["name"],
        description=postgres_fields["description"],
        type=postgres_fields["type"],
        software=postgres_fields["software"],
        team=postgres_fields["team"],
        tags=pb.Tags(tag=postgres_fields["tags"]),
        properties=postgres_fields["properties"],
        serialized_config=PostgresConfig(
            host=postgres_fields["host"],
            port=postgres_fields["port"],
            database=postgres_fields["database"],
            user=postgres_fields["user"],
            password=postgres_fields["password"],
            sslmode=postgres_fields["sslmode"],
        ).serialize(),
    )


def clickhouse_proto():
    return pb.Provider(
        name=clickhouse_fields["name"],
        description=clickhouse_fields["description"],
        type=clickhouse_fields["type"],
        software=clickhouse_fields["software"],
        team=clickhouse_fields["team"],
        tags=pb.Tags(tag=clickhouse_fields["tags"]),
        properties=clickhouse_fields["properties"],
        serialized_config=ClickHouseConfig(
            host=clickhouse_fields["host"],
            port=clickhouse_fields["port"],
            database=clickhouse_fields["database"],
            user=clickhouse_fields["user"],
            password=clickhouse_fields["password"],
            ssl=clickhouse_fields["sslmode"],
        ).serialize(),
    )


def redshift_proto():
    return pb.Provider(
        name=redshift_fields["name"],
        description=redshift_fields["description"],
        type=redshift_fields["type"],
        software=redshift_fields["software"],
        team=redshift_fields["team"],
        tags=pb.Tags(tag=redshift_fields["tags"]),
        properties=redshift_fields["properties"],
        serialized_config=RedshiftConfig(
            host=redshift_fields["host"],
            port=redshift_fields["port"],
            database=redshift_fields["database"],
            user=redshift_fields["user"],
            password=redshift_fields["password"],
            sslmode=redshift_fields["sslmode"],
        ).serialize(),
    )


def bigquery_proto():
    return pb.Provider(
        name=bigquery_fields["name"],
        description=bigquery_fields["description"],
        type=bigquery_fields["type"],
        team=bigquery_fields["team"],
        tags=pb.Tags(tag=bigquery_fields["tags"]),
        properties=bigquery_fields["properties"],
        serialized_config=BigQueryConfig(
            project_id=bigquery_fields["project_id"],
            dataset_id=bigquery_fields["dataset_id"],
            credentials=GCPCredentials(
                project_id=bigquery_fields["project_id"],
                credentials_path=bigquery_fields["credentials_path"],
            ),
        ).serialize(),
    )


def cassandra_proto():
    return pb.Provider(
        name=cassandra_fields["name"],
        description=cassandra_fields["description"],
        type=cassandra_fields["type"],
        team=cassandra_fields["team"],
        tags=pb.Tags(tag=cassandra_fields["tags"]),
        properties=cassandra_fields["properties"],
        serialized_config=CassandraConfig(
            host=cassandra_fields["host"],
            port=cassandra_fields["port"],
            username=cassandra_fields["username"],
            password=cassandra_fields["password"],
            keyspace=cassandra_fields["keyspace"],
            consistency=cassandra_fields["consistency"],
            replication=cassandra_fields["replication"],
        ).serialize(),
    )


def firestore_proto():
    return pb.Provider(
        name=firestore_fields["name"],
        description=firestore_fields["description"],
        type=firestore_fields["type"],
        team=firestore_fields["team"],
        tags=pb.Tags(tag=firestore_fields["tags"]),
        properties=firestore_fields["properties"],
        serialized_config=FirestoreConfig(
            collection=firestore_fields["collection"],
            project_id=firestore_fields["project_id"],
            credentials=GCPCredentials(
                project_id=firestore_fields["project_id"],
                credentials_path=firestore_fields["credentials_path"],
            ),
        ).serialize(),
    )


def dynamo_proto():
    return pb.Provider(
        name=dynamodb_fields["name"],
        description=dynamodb_fields["description"],
        type=dynamodb_fields["type"],
        team=dynamodb_fields["team"],
        tags=pb.Tags(tag=dynamodb_fields["tags"]),
        properties=dynamodb_fields["properties"],
        serialized_config=DynamodbConfig(
            credentials=AWSStaticCredentials(
                access_key=dynamodb_fields["aws_access_key_id"],
                secret_key=dynamodb_fields["aws_secret_access_key"],
            ),
            region=dynamodb_fields["region"],
            should_import_from_s3=dynamodb_fields["should_import_from_s3"],
        ).serialize(),
    )


def redis_proto():
    return pb.Provider(
        name=redis_fields["name"],
        description=redis_fields["description"],
        type=redis_fields["type"],
        team=redis_fields["team"],
        tags=pb.Tags(tag=redis_fields["tags"]),
        properties=redis_fields["properties"],
        serialized_config=RedisConfig(
            host=redis_fields["host"],
            port=redis_fields["port"],
            password=redis_fields["password"],
            db=redis_fields["database"],
        ).serialize(),
    )


def mongo_proto():
    return pb.Provider(
        name=mongodb_fields["name"],
        description=mongodb_fields["description"],
        type=mongodb_fields["type"],
        team=mongodb_fields["team"],
        tags=pb.Tags(tag=mongodb_fields["tags"]),
        properties=mongodb_fields["properties"],
        serialized_config=MongoDBConfig(
            username=mongodb_fields["username"],
            password=mongodb_fields["password"],
            database=mongodb_fields["database"],
            host=mongodb_fields["host"],
            port=mongodb_fields["port"],
            throughput=mongodb_fields["throughput"],
        ).serialize(),
    )


def snowflake_legacy_proto():
    return pb.Provider(
        name=snowflake_legacy_fields["name"],
        description=snowflake_legacy_fields["description"],
        type=snowflake_legacy_fields["type"],
        team=snowflake_legacy_fields["team"],
        tags=pb.Tags(tag=snowflake_legacy_fields["tags"]),
        properties=snowflake_legacy_fields["properties"],
        serialized_config=SnowflakeConfig(
            username=snowflake_legacy_fields["username"],
            password=snowflake_legacy_fields["password"],
            schema=snowflake_legacy_fields["schema"],
            database=snowflake_legacy_fields["database"],
            warehouse=snowflake_legacy_fields["warehouse"],
            role=snowflake_legacy_fields["role"],
            account_locator=snowflake_legacy_fields["account_locator"],
        ).serialize(),
    )


def snowflake_proto():
    return pb.Provider(
        name=snowflake_fields["name"],
        description=snowflake_fields["description"],
        type=snowflake_fields["type"],
        team=snowflake_fields["team"],
        tags=pb.Tags(tag=snowflake_fields["tags"]),
        properties=snowflake_fields["properties"],
        serialized_config=SnowflakeConfig(
            username=snowflake_fields["username"],
            password=snowflake_fields["password"],
            account=snowflake_fields["account"],
            schema=snowflake_fields["schema"],
            organization=snowflake_fields["organization"],
            database=snowflake_fields["database"],
            warehouse=snowflake_fields["warehouse"],
            role=snowflake_fields["role"],
        ).serialize(),
    )


def pinecone_proto():
    return pb.Provider(
        name=pinecone_fields["name"],
        description=pinecone_fields["description"],
        type=pinecone_fields["type"],
        team=pinecone_fields["team"],
        tags=pb.Tags(tag=pinecone_fields["tags"]),
        properties=pinecone_fields["properties"],
        serialized_config=PineconeConfig(
            project_id=pinecone_fields["project_id"],
            environment=pinecone_fields["environment"],
            api_key=pinecone_fields["api_key"],
        ).serialize(),
    )


def s3_proto():
    return pb.Provider(
        name=s3_fields["name"],
        description=s3_fields["description"],
        type="S3_ONLINE",
        team=s3_fields["team"],
        tags=pb.Tags(tag=s3_fields["tags"]),
        properties=s3_fields["properties"],
        serialized_config=S3StoreConfig(
            credentials=AWSStaticCredentials(
                access_key=s3_fields["access_key"],
                secret_key=s3_fields["secret_key"],
            ),
            bucket_region=s3_fields["bucket_region"],
            bucket_path=s3_fields["bucket_name"],
            path=s3_fields["path"],
        ).serialize(),
    )


def gcs_proto():
    return pb.Provider(
        name=gcs_fields["name"],
        description=gcs_fields["description"],
        type="GCS_ONLINE",
        team=gcs_fields["team"],
        tags=pb.Tags(tag=gcs_fields["tags"]),
        properties=gcs_fields["properties"],
        serialized_config=GCSFileStoreConfig(
            bucket_name=gcs_fields["bucket_name"],
            bucket_path=gcs_fields["bucket_path"],
            credentials=GCPCredentials(
                project_id=gcs_fields["project_id"],
                credentials_path=gcs_fields["credentials_path"],
            ),
        ).serialize(),
    )


def azure_proto():
    return pb.Provider(
        name=azure_fields["name"],
        description=azure_fields["description"],
        type="AZURE_ONLINE",
        team=azure_fields["team"],
        tags=pb.Tags(tag=azure_fields["tags"]),
        properties=azure_fields["properties"],
        serialized_config=AzureFileStoreConfig(
            account_name=azure_fields["account_name"],
            account_key=azure_fields["account_key"],
            container_name=azure_fields["container_name"],
            root_path=azure_fields["root_path"],
        ).serialize(),
    )


def hdfs_proto():
    return pb.Provider(
        name=hdfs_fields["name"],
        description=hdfs_fields["description"],
        type="HDFS_ONLINE",
        team=hdfs_fields["team"],
        tags=pb.Tags(tag=hdfs_fields["tags"]),
        properties=hdfs_fields["properties"],
        serialized_config=HDFSConfig(
            host=hdfs_fields["host"],
            port=hdfs_fields["port"],
            path=hdfs_fields["path"],
            hdfs_site_file=hdfs_fields["hdfs_site_file"],
            core_site_file=hdfs_fields["core_site_file"],
            credentials=BasicCredentials(
                username=hdfs_fields["username"], password=hdfs_fields["password"]
            ),
        ).serialize(),
    )


def spark_proto(name, executor, storage):
    return pb.Provider(
        name=name,
        description=spark_fields["description"],
        team=spark_fields["team"],
        tags=pb.Tags(tag=spark_fields["tags"]),
        properties=spark_fields["properties"],
        serialized_config=SparkConfig(
            executor_config=executor.config(),
            executor_type=executor.type(),
            store_config=storage.config(),
            store_type=storage.type(),
        ).serialize(),
    )


class MockStub:
    # Get Provider returns a proto
    def GetProviders(self, req):
        name = next(req).name
        if name == postgres_fields["name"]:
            yield postgres_proto()
        elif name == clickhouse_fields["name"]:
            yield clickhouse_proto()
        elif name == redshift_fields["name"]:
            yield redshift_proto()
        elif name == bigquery_fields["name"]:
            yield bigquery_proto()
        elif name == cassandra_fields["name"]:
            yield cassandra_proto()
        elif name == firestore_fields["name"]:
            yield firestore_proto()
        elif name == dynamodb_fields["name"]:
            yield dynamo_proto()
        elif name == redis_fields["name"]:
            yield redis_proto()
        elif name == mongodb_fields["name"]:
            yield mongo_proto()
        elif name == snowflake_fields["name"]:
            yield snowflake_proto()
        elif name == snowflake_legacy_fields["name"]:
            yield snowflake_legacy_proto()
        elif name == pinecone_fields["name"]:
            yield pinecone_proto()
        elif name == s3_fields["name"]:
            yield s3_proto()
        elif name == azure_fields["name"]:
            yield azure_proto()
        elif name == gcs_fields["name"]:
            yield gcs_proto()
        elif name == hdfs_fields["name"]:
            yield hdfs_proto()
        elif name == spark_fields["name_emr_s3"]:
            yield spark_proto(name, emr_executor, s3_store)
        elif name == spark_fields["name_emr_hdfs"]:
            yield spark_proto(name, emr_executor, hdfs_store)
        elif name == spark_fields["name_emr_gcs"]:
            yield spark_proto(name, emr_executor, gcs_store)
        elif name == spark_fields["name_emr_azure"]:
            yield spark_proto(name, emr_executor, azure_store)
        elif name == spark_fields["name_databricks_s3"]:
            yield spark_proto(name, databricks_executor, s3_store)
        elif name == spark_fields["name_databricks_hdfs"]:
            yield spark_proto(name, databricks_executor, hdfs_store)
        elif name == spark_fields["name_databricks_gcs"]:
            yield spark_proto(name, databricks_executor, gcs_store)
        elif name == spark_fields["name_databricks_azure"]:
            yield spark_proto(name, databricks_executor, azure_store)
        elif name == spark_fields["name_generic_s3"]:
            yield spark_proto(name, generic_executor, s3_store)
        elif name == spark_fields["name_generic_hdfs"]:
            yield spark_proto(name, generic_executor, hdfs_store)
        elif name == spark_fields["name_generic_gcs"]:
            yield spark_proto(name, generic_executor, gcs_store)
        elif name == spark_fields["name_generic_azure"]:
            yield spark_proto(name, generic_executor, azure_store)
        else:
            raise ValueError(f"Untested Provider {name}")


def test_get_postgres():
    client = ff.Client(host="localhost:7878", insecure=True, dry_run=True)
    client._stub = MockStub()

    expected = ff.register_postgres(
        name=postgres_fields["name"],
        host=postgres_fields["host"],
        port=postgres_fields["port"],
        user=postgres_fields["user"],
        password=postgres_fields["password"],
        database=postgres_fields["database"],
        description=postgres_fields["description"],
        team=postgres_fields["team"],
        sslmode=postgres_fields["sslmode"],
        tags=postgres_fields["tags"],
        properties=postgres_fields["properties"],
    )
    got = client.get_postgres(postgres_fields["name"])
    assert type(expected) == type(got)
    assert expected == got


def test_get_redshift():
    client = ff.Client(host="localhost:7878", insecure=True, dry_run=True)
    client._stub = MockStub()

    expected = ff.register_redshift(
        name=redshift_fields["name"],
        host=redshift_fields["host"],
        port=redshift_fields["port"],
        user=redshift_fields["user"],
        password=redshift_fields["password"],
        database=redshift_fields["database"],
        description=redshift_fields["description"],
        team=redshift_fields["team"],
        sslmode=redshift_fields["sslmode"],
        tags=redshift_fields["tags"],
        properties=redshift_fields["properties"],
    )
    got = client.get_redshift(redshift_fields["name"])
    assert type(expected) == type(got)
    assert expected == got


def test_get_bigquery():
    client = ff.Client(host="localhost:7878", insecure=True, dry_run=True)
    client._stub = MockStub()

    expected = ff.register_bigquery(
        name=bigquery_fields["name"],
        project_id=bigquery_fields["project_id"],
        dataset_id=bigquery_fields["dataset_id"],
        credentials=GCPCredentials(
            project_id=bigquery_fields["project_id"],
            credentials_path=bigquery_fields["credentials_path"],
        ),
        description=bigquery_fields["description"],
        team=bigquery_fields["team"],
        tags=bigquery_fields["tags"],
        properties=bigquery_fields["properties"],
    )
    got = client.get_bigquery(bigquery_fields["name"])
    assert type(expected) == type(got)
    assert expected == got


def test_get_cassandra():
    client = ff.Client(host="localhost:7878", insecure=True, dry_run=True)
    client._stub = MockStub()

    expected = ff.register_cassandra(
        name=cassandra_fields["name"],
        host=cassandra_fields["host"],
        port=cassandra_fields["port"],
        username=cassandra_fields["username"],
        password=cassandra_fields["password"],
        keyspace=cassandra_fields["keyspace"],
        consistency=cassandra_fields["consistency"],
        replication=cassandra_fields["replication"],
        description=cassandra_fields["description"],
        team=cassandra_fields["team"],
        tags=cassandra_fields["tags"],
        properties=cassandra_fields["properties"],
    )
    got = client.get_cassandra(cassandra_fields["name"])
    assert type(expected) == type(got)
    assert expected == got


def test_get_firestore():
    client = ff.Client(host="localhost:7878", insecure=True, dry_run=True)
    client._stub = MockStub()

    expected = ff.register_firestore(
        name=firestore_fields["name"],
        collection=firestore_fields["collection"],
        project_id=firestore_fields["project_id"],
        credentials=GCPCredentials(
            project_id=firestore_fields["project_id"],
            credentials_path=firestore_fields["credentials_path"],
        ),
        description=firestore_fields["description"],
        team=firestore_fields["team"],
        tags=firestore_fields["tags"],
        properties=firestore_fields["properties"],
    )
    got = client.get_firestore(firestore_fields["name"])
    assert type(expected) == type(got)
    assert expected == got


def test_get_dynamodb():
    client = ff.Client(host="localhost:7878", insecure=True, dry_run=True)
    client._stub = MockStub()

    expected = ff.register_dynamodb(
        name=dynamodb_fields["name"],
        credentials=AWSStaticCredentials(
            access_key=dynamodb_fields["aws_access_key_id"],
            secret_key=dynamodb_fields["aws_secret_access_key"],
        ),
        region=dynamodb_fields["region"],
        should_import_from_s3=dynamodb_fields["should_import_from_s3"],
        description=dynamodb_fields["description"],
        team=dynamodb_fields["team"],
        tags=dynamodb_fields["tags"],
        properties=dynamodb_fields["properties"],
    )
    got = client.get_dynamodb(dynamodb_fields["name"])
    assert type(expected) == type(got)
    assert expected == got


def test_get_redis():
    client = ff.Client(host="localhost:7878", insecure=True, dry_run=True)
    client._stub = MockStub()

    expected = ff.register_redis(
        name=redis_fields["name"],
        host=redis_fields["host"],
        port=redis_fields["port"],
        password=redis_fields["password"],
        db=redis_fields["database"],
        description=redis_fields["description"],
        team=redis_fields["team"],
        tags=redis_fields["tags"],
        properties=redis_fields["properties"],
    )
    got = client.get_redis(redis_fields["name"])
    assert type(expected) == type(got)
    assert expected == got


def test_get_mongodb():
    client = ff.Client(host="localhost:7878", insecure=True, dry_run=True)
    client._stub = MockStub()

    expected = ff.register_mongodb(
        name=mongodb_fields["name"],
        username=mongodb_fields["username"],
        password=mongodb_fields["password"],
        database=mongodb_fields["database"],
        host=mongodb_fields["host"],
        port=mongodb_fields["port"],
        throughput=mongodb_fields["throughput"],
        description=mongodb_fields["description"],
        team=mongodb_fields["team"],
        tags=mongodb_fields["tags"],
        properties=mongodb_fields["properties"],
    )
    got = client.get_mongodb(mongodb_fields["name"])
    assert type(expected) == type(got)
    assert expected == got


def test_get_pinecone():
    client = ff.Client(host="localhost:7878", insecure=True, dry_run=True)
    client._stub = MockStub()

    expected = ff.register_pinecone(
        name=pinecone_fields["name"],
        project_id=pinecone_fields["project_id"],
        environment=pinecone_fields["environment"],
        api_key=pinecone_fields["api_key"],
        description=pinecone_fields["description"],
        team=pinecone_fields["team"],
        tags=pinecone_fields["tags"],
        properties=pinecone_fields["properties"],
    )
    got = client.get_pinecone(pinecone_fields["name"])
    assert type(expected) == type(got)
    assert expected == got


def test_get_snowflake():
    client = ff.Client(host="localhost:7878", insecure=True, dry_run=True)
    client._stub = MockStub()

    expected = ff.register_snowflake(
        name=snowflake_fields["name"],
        username=snowflake_fields["username"],
        password=snowflake_fields["password"],
        account=snowflake_fields["account"],
        organization=snowflake_fields["organization"],
        database=snowflake_fields["database"],
        warehouse=snowflake_fields["warehouse"],
        description=snowflake_fields["description"],
        role=snowflake_fields["role"],
        team=snowflake_fields["team"],
        tags=snowflake_fields["tags"],
        properties=snowflake_fields["properties"],
    )
    got = client.get_snowflake(snowflake_fields["name"])
    assert type(expected) == type(got)
    assert expected == got


def test_get_snowflake_legacy():
    client = ff.Client(host="localhost:7878", insecure=True, dry_run=True)
    client._stub = MockStub()

    expected = ff.register_snowflake_legacy(
        name=snowflake_legacy_fields["name"],
        username=snowflake_legacy_fields["username"],
        password=snowflake_legacy_fields["password"],
        account_locator=snowflake_legacy_fields["account_locator"],
        database=snowflake_legacy_fields["database"],
        warehouse=snowflake_legacy_fields["warehouse"],
        description=snowflake_legacy_fields["description"],
        role=snowflake_legacy_fields["role"],
        team=snowflake_legacy_fields["team"],
        tags=snowflake_legacy_fields["tags"],
        properties=snowflake_legacy_fields["properties"],
    )
    got = client.get_snowflake_legacy(snowflake_legacy_fields["name"])
    assert type(expected) == type(got)
    assert expected == got


def test_get_s3():
    client = ff.Client(host="localhost:7878", insecure=True, dry_run=True)
    client._stub = MockStub()

    credentials = AWSStaticCredentials(
        access_key=s3_fields["access_key"],
        secret_key=s3_fields["secret_key"],
    )

    # Returns a FileStoreProvider
    expected = ff.register_s3(
        name=s3_fields["name"],
        credentials=credentials,
        bucket_region=s3_fields["bucket_region"],
        bucket_name=s3_fields["bucket_name"],
        path=s3_fields["path"],
        description=s3_fields["description"],
        team=s3_fields["team"],
        tags=s3_fields["tags"],
        properties=s3_fields["properties"],
    )
    got = client.get_s3(s3_fields["name"])
    assert type(expected) == type(got)
    assert expected == got


def test_get_azure():
    client = ff.Client(host="localhost:7878", insecure=True, dry_run=True)
    client._stub = MockStub()

    # Returns a FileStoreProvider
    expected = ff.register_blob_store(
        name=azure_fields["name"],
        account_name=azure_fields["account_name"],
        account_key=azure_fields["account_key"],
        container_name=azure_fields["container_name"],
        root_path=azure_fields["root_path"],
        description=azure_fields["description"],
        team=azure_fields["team"],
        tags=azure_fields["tags"],
        properties=azure_fields["properties"],
    )
    got = client.get_blob_store(azure_fields["name"])
    assert type(expected) == type(got)
    assert expected == got


def test_get_gcs():
    client = ff.Client(host="localhost:7878", insecure=True, dry_run=True)
    client._stub = MockStub()

    # Returns a FileStoreProvider
    expected = ff.register_gcs(
        name=gcs_fields["name"],
        bucket_name=gcs_fields["bucket_name"],
        root_path=gcs_fields["bucket_path"],
        credentials=GCPCredentials(
            project_id=gcs_fields["project_id"],
            credentials_path=gcs_fields["credentials_path"],
        ),
        description=gcs_fields["description"],
        team=gcs_fields["team"],
        tags=gcs_fields["tags"],
        properties=gcs_fields["properties"],
    )
    got = client.get_gcs(gcs_fields["name"])
    assert type(expected) == type(got)
    assert expected == got


def test_get_hdfs():
    client = ff.Client(host="localhost:7878", insecure=True, dry_run=True)
    client._stub = MockStub()

    # Returns a FileStoreProvider
    expected = ff.register_hdfs(
        name=hdfs_fields["name"],
        host=hdfs_fields["host"],
        port=hdfs_fields["port"],
        path=hdfs_fields["path"],
        hdfs_site_file=hdfs_fields["hdfs_site_file"],
        core_site_file=hdfs_fields["core_site_file"],
        credentials=BasicCredentials(
            username=hdfs_fields["username"],
            password=hdfs_fields["password"],
        ),
        description=hdfs_fields["description"],
        team=hdfs_fields["team"],
        tags=hdfs_fields["tags"],
        properties=hdfs_fields["properties"],
    )
    got = client.get_hdfs(hdfs_fields["name"])
    assert type(expected) == type(got)
    assert expected == got


def test_get_spark_emr_s3():
    client = ff.Client(host="localhost:7878", insecure=True, dry_run=True)
    client._stub = MockStub()

    expected = ff.register_spark(
        name=spark_fields["name_emr_s3"],
        executor=emr_executor,
        filestore=client.get_s3(s3_fields["name"]),
        description=spark_fields["description"],
        team=spark_fields["team"],
        tags=spark_fields["tags"],
        properties=spark_fields["properties"],
    )
    got = client.get_spark(spark_fields["name_emr_s3"])
    assert type(expected) == type(got)
    assert expected == got


def test_get_spark_emr_hdfs():
    client = ff.Client(host="localhost:7878", insecure=True, dry_run=True)
    client._stub = MockStub()

    expected = ff.register_spark(
        name=spark_fields["name_emr_hdfs"],
        executor=emr_executor,
        filestore=client.get_hdfs(hdfs_fields["name"]),
        description=spark_fields["description"],
        team=spark_fields["team"],
        tags=spark_fields["tags"],
        properties=spark_fields["properties"],
    )
    got = client.get_spark(spark_fields["name_emr_hdfs"])
    assert type(expected) == type(got)
    assert expected == got


def test_get_spark_emr_gcs():
    client = ff.Client(host="localhost:7878", insecure=True, dry_run=True)
    client._stub = MockStub()

    expected = ff.register_spark(
        name=spark_fields["name_emr_gcs"],
        executor=emr_executor,
        filestore=client.get_gcs(gcs_fields["name"]),
        description=spark_fields["description"],
        team=spark_fields["team"],
        tags=spark_fields["tags"],
        properties=spark_fields["properties"],
    )
    got = client.get_spark(spark_fields["name_emr_gcs"])
    assert type(expected) == type(got)
    assert expected == got


def test_get_spark_emr_azure():
    client = ff.Client(host="localhost:7878", insecure=True, dry_run=True)
    client._stub = MockStub()

    expected = ff.register_spark(
        name=spark_fields["name_emr_azure"],
        executor=emr_executor,
        filestore=client.get_blob_store(azure_fields["name"]),
        description=spark_fields["description"],
        team=spark_fields["team"],
        tags=spark_fields["tags"],
        properties=spark_fields["properties"],
    )
    got = client.get_spark(spark_fields["name_emr_azure"])
    assert type(expected) == type(got)
    assert expected == got


def test_get_spark_databricks_s3():
    client = ff.Client(host="localhost:7878", insecure=True, dry_run=True)
    client._stub = MockStub()

    expected = ff.register_spark(
        name=spark_fields["name_databricks_s3"],
        executor=databricks_executor,
        filestore=client.get_s3(s3_fields["name"]),
        description=spark_fields["description"],
        team=spark_fields["team"],
        tags=spark_fields["tags"],
        properties=spark_fields["properties"],
    )
    got = client.get_spark(spark_fields["name_databricks_s3"])
    assert type(expected) == type(got)
    assert expected == got


def test_get_spark_databricks_hdfs():
    client = ff.Client(host="localhost:7878", insecure=True, dry_run=True)
    client._stub = MockStub()

    expected = ff.register_spark(
        name=spark_fields["name_databricks_hdfs"],
        executor=databricks_executor,
        filestore=client.get_hdfs(hdfs_fields["name"]),
        description=spark_fields["description"],
        team=spark_fields["team"],
        tags=spark_fields["tags"],
        properties=spark_fields["properties"],
    )
    got = client.get_spark(spark_fields["name_databricks_hdfs"])
    assert type(expected) == type(got)
    assert expected == got


def test_get_spark_databricks_gcs():
    client = ff.Client(host="localhost:7878", insecure=True, dry_run=True)
    client._stub = MockStub()

    expected = ff.register_spark(
        name=spark_fields["name_databricks_gcs"],
        executor=databricks_executor,
        filestore=client.get_gcs(gcs_fields["name"]),
        description=spark_fields["description"],
        team=spark_fields["team"],
        tags=spark_fields["tags"],
        properties=spark_fields["properties"],
    )
    got = client.get_spark(spark_fields["name_databricks_gcs"])
    assert type(expected) == type(got)
    assert expected == got


def test_get_spark_databricks_azure():
    client = ff.Client(host="localhost:7878", insecure=True, dry_run=True)
    client._stub = MockStub()

    expected = ff.register_spark(
        name=spark_fields["name_databricks_azure"],
        executor=databricks_executor,
        filestore=client.get_blob_store(azure_fields["name"]),
        description=spark_fields["description"],
        team=spark_fields["team"],
        tags=spark_fields["tags"],
        properties=spark_fields["properties"],
    )
    got = client.get_spark(spark_fields["name_databricks_azure"])
    assert type(expected) == type(got)
    assert expected == got


def test_get_spark_generic_s3():
    client = ff.Client(host="localhost:7878", insecure=True, dry_run=True)
    client._stub = MockStub()

    expected = ff.register_spark(
        name=spark_fields["name_generic_s3"],
        executor=generic_executor,
        filestore=client.get_s3(s3_fields["name"]),
        description=spark_fields["description"],
        team=spark_fields["team"],
        tags=spark_fields["tags"],
        properties=spark_fields["properties"],
    )
    got = client.get_spark(spark_fields["name_generic_s3"])
    assert type(expected) == type(got)
    assert expected == got


def test_get_spark_generic_hdfs():
    client = ff.Client(host="localhost:7878", insecure=True, dry_run=True)
    client._stub = MockStub()

    expected = ff.register_spark(
        name=spark_fields["name_generic_hdfs"],
        executor=generic_executor,
        filestore=client.get_hdfs(hdfs_fields["name"]),
        description=spark_fields["description"],
        team=spark_fields["team"],
        tags=spark_fields["tags"],
        properties=spark_fields["properties"],
    )
    got = client.get_spark(spark_fields["name_generic_hdfs"])
    assert type(expected) == type(got)
    assert expected == got


def test_get_spark_generic_gcs():
    client = ff.Client(host="localhost:7878", insecure=True, dry_run=True)
    client._stub = MockStub()

    expected = ff.register_spark(
        name=spark_fields["name_generic_gcs"],
        executor=generic_executor,
        filestore=client.get_gcs(gcs_fields["name"]),
        description=spark_fields["description"],
        team=spark_fields["team"],
        tags=spark_fields["tags"],
        properties=spark_fields["properties"],
    )
    got = client.get_spark(spark_fields["name_generic_gcs"])
    assert type(expected) == type(got)
    assert expected == got


def test_get_spark_generic_azure():
    client = ff.Client(host="localhost:7878", insecure=True, dry_run=True)
    client._stub = MockStub()

    expected = ff.register_spark(
        name=spark_fields["name_generic_azure"],
        executor=generic_executor,
        filestore=client.get_blob_store(azure_fields["name"]),
        description=spark_fields["description"],
        team=spark_fields["team"],
        tags=spark_fields["tags"],
        properties=spark_fields["properties"],
    )
    got = client.get_spark(spark_fields["name_generic_azure"])
    assert type(expected) == type(got)
    assert expected == got


def test_clickhouse():
    client = ff.Client(host="localhost:7878", insecure=True, dry_run=True)
    client._stub = MockStub()

    expected = ff.register_clickhouse(
        name=clickhouse_fields["name"],
        host=clickhouse_fields["host"],
        port=clickhouse_fields["port"],
        user=clickhouse_fields["user"],
        password=clickhouse_fields["password"],
        database=clickhouse_fields["database"],
        description=clickhouse_fields["description"],
        team=clickhouse_fields["team"],
        ssl=clickhouse_fields["sslmode"],
        tags=clickhouse_fields["tags"],
        properties=clickhouse_fields["properties"],
    )
    got = client.get_clickhouse(clickhouse_fields["name"])
    assert type(expected) == type(got)
    assert expected == got



import pkg_resources
import requests
import threading

version_check_url = "https://version.featureform.com"


def get_package_version():
    return pkg_resources.get_distribution("featureform-enterprise").version


def check_up_to_date(local, client):
    download_thread = threading.Thread(
        target=run_version_check, name="Downloader", args=(local, client)
    )
    download_thread.start()


def run_version_check(local, client):
    try:
        version = get_package_version()
        requests.get(
            version_check_url,
            params={"local": local, "client": client, "version": version},
        )
    except:
        pass



"""
@generated by mypy-protobuf.  Do not edit manually!
isort:skip_file
This Source Code Form is subject to the terms of the Mozilla Public
License, v. 2.0. If a copy of the MPL was not distributed with this
file, You can obtain one at https://mozilla.org/MPL/2.0/.
"""

import builtins
import collections.abc
import google.protobuf.descriptor
import google.protobuf.internal.containers
import google.protobuf.internal.enum_type_wrapper
import google.protobuf.message
import sys
import typing

if sys.version_info >= (3, 10):
    import typing as typing_extensions
else:
    import typing_extensions

DESCRIPTOR: google.protobuf.descriptor.FileDescriptor

class _RequestType:
    ValueType = typing.NewType("ValueType", builtins.int)
    V: typing_extensions.TypeAlias = ValueType

class _RequestTypeEnumTypeWrapper(google.protobuf.internal.enum_type_wrapper._EnumTypeWrapper[_RequestType.ValueType], builtins.type):
    DESCRIPTOR: google.protobuf.descriptor.EnumDescriptor
    INITIALIZE: _RequestType.ValueType  # 0
    TRAINING: _RequestType.ValueType  # 1
    """Client is requesting training data"""
    TEST: _RequestType.ValueType  # 2
    """Client is requesting test data"""

class RequestType(_RequestType, metaclass=_RequestTypeEnumTypeWrapper): ...

INITIALIZE: RequestType.ValueType  # 0
TRAINING: RequestType.ValueType  # 1
"""Client is requesting training data"""
TEST: RequestType.ValueType  # 2
"""Client is requesting test data"""
global___RequestType = RequestType

@typing.final
class Model(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    NAME_FIELD_NUMBER: builtins.int
    name: builtins.str
    def __init__(
        self,
        *,
        name: builtins.str = ...,
    ) -> None: ...
    def ClearField(self, field_name: typing.Literal["name", b"name"]) -> None: ...

global___Model = Model

@typing.final
class TrainingDataRequest(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    ID_FIELD_NUMBER: builtins.int
    MODEL_FIELD_NUMBER: builtins.int
    @property
    def id(self) -> global___TrainingDataID: ...
    @property
    def model(self) -> global___Model: ...
    def __init__(
        self,
        *,
        id: global___TrainingDataID | None = ...,
        model: global___Model | None = ...,
    ) -> None: ...
    def HasField(self, field_name: typing.Literal["id", b"id", "model", b"model"]) -> builtins.bool: ...
    def ClearField(self, field_name: typing.Literal["id", b"id", "model", b"model"]) -> None: ...

global___TrainingDataRequest = TrainingDataRequest

@typing.final
class TrainingDataID(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    NAME_FIELD_NUMBER: builtins.int
    VERSION_FIELD_NUMBER: builtins.int
    name: builtins.str
    version: builtins.str
    def __init__(
        self,
        *,
        name: builtins.str = ...,
        version: builtins.str = ...,
    ) -> None: ...
    def ClearField(self, field_name: typing.Literal["name", b"name", "version", b"version"]) -> None: ...

global___TrainingDataID = TrainingDataID

@typing.final
class TrainingDataRow(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    FEATURES_FIELD_NUMBER: builtins.int
    LABEL_FIELD_NUMBER: builtins.int
    @property
    def features(self) -> google.protobuf.internal.containers.RepeatedCompositeFieldContainer[global___Value]: ...
    @property
    def label(self) -> global___Value: ...
    def __init__(
        self,
        *,
        features: collections.abc.Iterable[global___Value] | None = ...,
        label: global___Value | None = ...,
    ) -> None: ...
    def HasField(self, field_name: typing.Literal["label", b"label"]) -> builtins.bool: ...
    def ClearField(self, field_name: typing.Literal["features", b"features", "label", b"label"]) -> None: ...

global___TrainingDataRow = TrainingDataRow

@typing.final
class FeatureServeRequest(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    FEATURES_FIELD_NUMBER: builtins.int
    ENTITIES_FIELD_NUMBER: builtins.int
    MODEL_FIELD_NUMBER: builtins.int
    @property
    def features(self) -> google.protobuf.internal.containers.RepeatedCompositeFieldContainer[global___FeatureID]: ...
    @property
    def entities(self) -> google.protobuf.internal.containers.RepeatedCompositeFieldContainer[global___Entity]: ...
    @property
    def model(self) -> global___Model: ...
    def __init__(
        self,
        *,
        features: collections.abc.Iterable[global___FeatureID] | None = ...,
        entities: collections.abc.Iterable[global___Entity] | None = ...,
        model: global___Model | None = ...,
    ) -> None: ...
    def HasField(self, field_name: typing.Literal["model", b"model"]) -> builtins.bool: ...
    def ClearField(self, field_name: typing.Literal["entities", b"entities", "features", b"features", "model", b"model"]) -> None: ...

global___FeatureServeRequest = FeatureServeRequest

@typing.final
class FeatureRow(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    VALUES_FIELD_NUMBER: builtins.int
    VALUE_LISTS_FIELD_NUMBER: builtins.int
    @property
    def values(self) -> google.protobuf.internal.containers.RepeatedCompositeFieldContainer[global___Value]:
        """Deprecated: Use `value_lists` instead."""

    @property
    def value_lists(self) -> google.protobuf.internal.containers.RepeatedCompositeFieldContainer[global___ValueList]: ...
    def __init__(
        self,
        *,
        values: collections.abc.Iterable[global___Value] | None = ...,
        value_lists: collections.abc.Iterable[global___ValueList] | None = ...,
    ) -> None: ...
    def ClearField(self, field_name: typing.Literal["value_lists", b"value_lists", "values", b"values"]) -> None: ...

global___FeatureRow = FeatureRow

@typing.final
class ValueList(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    VALUES_FIELD_NUMBER: builtins.int
    @property
    def values(self) -> google.protobuf.internal.containers.RepeatedCompositeFieldContainer[global___Value]: ...
    def __init__(
        self,
        *,
        values: collections.abc.Iterable[global___Value] | None = ...,
    ) -> None: ...
    def ClearField(self, field_name: typing.Literal["values", b"values"]) -> None: ...

global___ValueList = ValueList

@typing.final
class BatchFeatureServeRequest(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    FEATURES_FIELD_NUMBER: builtins.int
    @property
    def features(self) -> google.protobuf.internal.containers.RepeatedCompositeFieldContainer[global___FeatureID]: ...
    def __init__(
        self,
        *,
        features: collections.abc.Iterable[global___FeatureID] | None = ...,
    ) -> None: ...
    def ClearField(self, field_name: typing.Literal["features", b"features"]) -> None: ...

global___BatchFeatureServeRequest = BatchFeatureServeRequest

@typing.final
class BatchFeatureRow(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    ENTITY_FIELD_NUMBER: builtins.int
    FEATURES_FIELD_NUMBER: builtins.int
    @property
    def entity(self) -> global___Value: ...
    @property
    def features(self) -> google.protobuf.internal.containers.RepeatedCompositeFieldContainer[global___Value]: ...
    def __init__(
        self,
        *,
        entity: global___Value | None = ...,
        features: collections.abc.Iterable[global___Value] | None = ...,
    ) -> None: ...
    def HasField(self, field_name: typing.Literal["entity", b"entity"]) -> builtins.bool: ...
    def ClearField(self, field_name: typing.Literal["entity", b"entity", "features", b"features"]) -> None: ...

global___BatchFeatureRow = BatchFeatureRow

@typing.final
class FeatureID(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    NAME_FIELD_NUMBER: builtins.int
    VERSION_FIELD_NUMBER: builtins.int
    name: builtins.str
    version: builtins.str
    def __init__(
        self,
        *,
        name: builtins.str = ...,
        version: builtins.str = ...,
    ) -> None: ...
    def ClearField(self, field_name: typing.Literal["name", b"name", "version", b"version"]) -> None: ...

global___FeatureID = FeatureID

@typing.final
class Entity(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    NAME_FIELD_NUMBER: builtins.int
    VALUE_FIELD_NUMBER: builtins.int
    VALUES_FIELD_NUMBER: builtins.int
    name: builtins.str
    value: builtins.str
    """Deprecated: Use `values` instead."""
    @property
    def values(self) -> google.protobuf.internal.containers.RepeatedScalarFieldContainer[builtins.str]: ...
    def __init__(
        self,
        *,
        name: builtins.str = ...,
        value: builtins.str = ...,
        values: collections.abc.Iterable[builtins.str] | None = ...,
    ) -> None: ...
    def ClearField(self, field_name: typing.Literal["name", b"name", "value", b"value", "values", b"values"]) -> None: ...

global___Entity = Entity

@typing.final
class Value(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    STR_VALUE_FIELD_NUMBER: builtins.int
    INT_VALUE_FIELD_NUMBER: builtins.int
    FLOAT_VALUE_FIELD_NUMBER: builtins.int
    DOUBLE_VALUE_FIELD_NUMBER: builtins.int
    INT64_VALUE_FIELD_NUMBER: builtins.int
    INT32_VALUE_FIELD_NUMBER: builtins.int
    BOOL_VALUE_FIELD_NUMBER: builtins.int
    ON_DEMAND_FUNCTION_FIELD_NUMBER: builtins.int
    VECTOR32_VALUE_FIELD_NUMBER: builtins.int
    UINT32_VALUE_FIELD_NUMBER: builtins.int
    UINT64_VALUE_FIELD_NUMBER: builtins.int
    str_value: builtins.str
    int_value: builtins.int
    float_value: builtins.float
    double_value: builtins.float
    int64_value: builtins.int
    int32_value: builtins.int
    bool_value: builtins.bool
    on_demand_function: builtins.bytes
    uint32_value: builtins.int
    uint64_value: builtins.int
    @property
    def vector32_value(self) -> global___Vector32: ...
    def __init__(
        self,
        *,
        str_value: builtins.str = ...,
        int_value: builtins.int = ...,
        float_value: builtins.float = ...,
        double_value: builtins.float = ...,
        int64_value: builtins.int = ...,
        int32_value: builtins.int = ...,
        bool_value: builtins.bool = ...,
        on_demand_function: builtins.bytes = ...,
        vector32_value: global___Vector32 | None = ...,
        uint32_value: builtins.int = ...,
        uint64_value: builtins.int = ...,
    ) -> None: ...
    def HasField(self, field_name: typing.Literal["bool_value", b"bool_value", "double_value", b"double_value", "float_value", b"float_value", "int32_value", b"int32_value", "int64_value", b"int64_value", "int_value", b"int_value", "on_demand_function", b"on_demand_function", "str_value", b"str_value", "uint32_value", b"uint32_value", "uint64_value", b"uint64_value", "value", b"value", "vector32_value", b"vector32_value"]) -> builtins.bool: ...
    def ClearField(self, field_name: typing.Literal["bool_value", b"bool_value", "double_value", b"double_value", "float_value", b"float_value", "int32_value", b"int32_value", "int64_value", b"int64_value", "int_value", b"int_value", "on_demand_function", b"on_demand_function", "str_value", b"str_value", "uint32_value", b"uint32_value", "uint64_value", b"uint64_value", "value", b"value", "vector32_value", b"vector32_value"]) -> None: ...
    def WhichOneof(self, oneof_group: typing.Literal["value", b"value"]) -> typing.Literal["str_value", "int_value", "float_value", "double_value", "int64_value", "int32_value", "bool_value", "on_demand_function", "vector32_value", "uint32_value", "uint64_value"] | None: ...

global___Value = Value

@typing.final
class SourceID(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    NAME_FIELD_NUMBER: builtins.int
    VERSION_FIELD_NUMBER: builtins.int
    name: builtins.str
    version: builtins.str
    def __init__(
        self,
        *,
        name: builtins.str = ...,
        version: builtins.str = ...,
    ) -> None: ...
    def ClearField(self, field_name: typing.Literal["name", b"name", "version", b"version"]) -> None: ...

global___SourceID = SourceID

@typing.final
class SourceDataRequest(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    ID_FIELD_NUMBER: builtins.int
    LIMIT_FIELD_NUMBER: builtins.int
    limit: builtins.int
    @property
    def id(self) -> global___SourceID: ...
    def __init__(
        self,
        *,
        id: global___SourceID | None = ...,
        limit: builtins.int = ...,
    ) -> None: ...
    def HasField(self, field_name: typing.Literal["id", b"id"]) -> builtins.bool: ...
    def ClearField(self, field_name: typing.Literal["id", b"id", "limit", b"limit"]) -> None: ...

global___SourceDataRequest = SourceDataRequest

@typing.final
class SourceColumnRequest(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    ID_FIELD_NUMBER: builtins.int
    @property
    def id(self) -> global___SourceID: ...
    def __init__(
        self,
        *,
        id: global___SourceID | None = ...,
    ) -> None: ...
    def HasField(self, field_name: typing.Literal["id", b"id"]) -> builtins.bool: ...
    def ClearField(self, field_name: typing.Literal["id", b"id"]) -> None: ...

global___SourceColumnRequest = SourceColumnRequest

@typing.final
class SourceDataRow(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    ROWS_FIELD_NUMBER: builtins.int
    @property
    def rows(self) -> google.protobuf.internal.containers.RepeatedCompositeFieldContainer[global___Value]: ...
    def __init__(
        self,
        *,
        rows: collections.abc.Iterable[global___Value] | None = ...,
    ) -> None: ...
    def ClearField(self, field_name: typing.Literal["rows", b"rows"]) -> None: ...

global___SourceDataRow = SourceDataRow

@typing.final
class SourceDataColumns(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    COLUMNS_FIELD_NUMBER: builtins.int
    @property
    def columns(self) -> google.protobuf.internal.containers.RepeatedScalarFieldContainer[builtins.str]: ...
    def __init__(
        self,
        *,
        columns: collections.abc.Iterable[builtins.str] | None = ...,
    ) -> None: ...
    def ClearField(self, field_name: typing.Literal["columns", b"columns"]) -> None: ...

global___SourceDataColumns = SourceDataColumns

@typing.final
class TrainingDataColumnsRequest(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    ID_FIELD_NUMBER: builtins.int
    @property
    def id(self) -> global___TrainingDataID: ...
    def __init__(
        self,
        *,
        id: global___TrainingDataID | None = ...,
    ) -> None: ...
    def HasField(self, field_name: typing.Literal["id", b"id"]) -> builtins.bool: ...
    def ClearField(self, field_name: typing.Literal["id", b"id"]) -> None: ...

global___TrainingDataColumnsRequest = TrainingDataColumnsRequest

@typing.final
class TrainingColumns(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    FEATURES_FIELD_NUMBER: builtins.int
    LABEL_FIELD_NUMBER: builtins.int
    label: builtins.str
    @property
    def features(self) -> google.protobuf.internal.containers.RepeatedScalarFieldContainer[builtins.str]: ...
    def __init__(
        self,
        *,
        features: collections.abc.Iterable[builtins.str] | None = ...,
        label: builtins.str = ...,
    ) -> None: ...
    def ClearField(self, field_name: typing.Literal["features", b"features", "label", b"label"]) -> None: ...

global___TrainingColumns = TrainingColumns

@typing.final
class Vector32(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    VALUE_FIELD_NUMBER: builtins.int
    @property
    def value(self) -> google.protobuf.internal.containers.RepeatedScalarFieldContainer[builtins.float]: ...
    def __init__(
        self,
        *,
        value: collections.abc.Iterable[builtins.float] | None = ...,
    ) -> None: ...
    def ClearField(self, field_name: typing.Literal["value", b"value"]) -> None: ...

global___Vector32 = Vector32

@typing.final
class NearestRequest(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    ID_FIELD_NUMBER: builtins.int
    VECTOR_FIELD_NUMBER: builtins.int
    K_FIELD_NUMBER: builtins.int
    k: builtins.int
    @property
    def id(self) -> global___FeatureID: ...
    @property
    def vector(self) -> global___Vector32: ...
    def __init__(
        self,
        *,
        id: global___FeatureID | None = ...,
        vector: global___Vector32 | None = ...,
        k: builtins.int = ...,
    ) -> None: ...
    def HasField(self, field_name: typing.Literal["id", b"id", "vector", b"vector"]) -> builtins.bool: ...
    def ClearField(self, field_name: typing.Literal["id", b"id", "k", b"k", "vector", b"vector"]) -> None: ...

global___NearestRequest = NearestRequest

@typing.final
class NearestResponse(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    ENTITIES_FIELD_NUMBER: builtins.int
    @property
    def entities(self) -> google.protobuf.internal.containers.RepeatedScalarFieldContainer[builtins.str]: ...
    def __init__(
        self,
        *,
        entities: collections.abc.Iterable[builtins.str] | None = ...,
    ) -> None: ...
    def ClearField(self, field_name: typing.Literal["entities", b"entities"]) -> None: ...

global___NearestResponse = NearestResponse

@typing.final
class ResourceIdRequest(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    NAME_FIELD_NUMBER: builtins.int
    VARIANT_FIELD_NUMBER: builtins.int
    TYPE_FIELD_NUMBER: builtins.int
    name: builtins.str
    variant: builtins.str
    type: builtins.int
    def __init__(
        self,
        *,
        name: builtins.str = ...,
        variant: builtins.str = ...,
        type: builtins.int = ...,
    ) -> None: ...
    def ClearField(self, field_name: typing.Literal["name", b"name", "type", b"type", "variant", b"variant"]) -> None: ...

global___ResourceIdRequest = ResourceIdRequest

@typing.final
class ResourceLocation(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    LOCATION_FIELD_NUMBER: builtins.int
    location: builtins.str
    def __init__(
        self,
        *,
        location: builtins.str = ...,
    ) -> None: ...
    def ClearField(self, field_name: typing.Literal["location", b"location"]) -> None: ...

global___ResourceLocation = ResourceLocation

@typing.final
class TrainTestSplitRequest(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    ID_FIELD_NUMBER: builtins.int
    MODEL_FIELD_NUMBER: builtins.int
    TEST_SIZE_FIELD_NUMBER: builtins.int
    TRAIN_SIZE_FIELD_NUMBER: builtins.int
    SHUFFLE_FIELD_NUMBER: builtins.int
    RANDOM_STATE_FIELD_NUMBER: builtins.int
    REQUEST_TYPE_FIELD_NUMBER: builtins.int
    BATCH_SIZE_FIELD_NUMBER: builtins.int
    test_size: builtins.float
    train_size: builtins.float
    shuffle: builtins.bool
    random_state: builtins.int
    request_type: global___RequestType.ValueType
    batch_size: builtins.int
    @property
    def id(self) -> global___TrainingDataID: ...
    @property
    def model(self) -> global___Model: ...
    def __init__(
        self,
        *,
        id: global___TrainingDataID | None = ...,
        model: global___Model | None = ...,
        test_size: builtins.float = ...,
        train_size: builtins.float = ...,
        shuffle: builtins.bool = ...,
        random_state: builtins.int = ...,
        request_type: global___RequestType.ValueType = ...,
        batch_size: builtins.int = ...,
    ) -> None: ...
    def HasField(self, field_name: typing.Literal["id", b"id", "model", b"model"]) -> builtins.bool: ...
    def ClearField(self, field_name: typing.Literal["batch_size", b"batch_size", "id", b"id", "model", b"model", "random_state", b"random_state", "request_type", b"request_type", "shuffle", b"shuffle", "test_size", b"test_size", "train_size", b"train_size"]) -> None: ...

global___TrainTestSplitRequest = TrainTestSplitRequest

@typing.final
class BatchTrainTestSplitResponse(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    REQUEST_TYPE_FIELD_NUMBER: builtins.int
    ITERATOR_DONE_FIELD_NUMBER: builtins.int
    INITIALIZED_FIELD_NUMBER: builtins.int
    DATA_FIELD_NUMBER: builtins.int
    request_type: global___RequestType.ValueType
    iterator_done: builtins.bool
    initialized: builtins.bool
    @property
    def data(self) -> global___TrainingDataRows: ...
    def __init__(
        self,
        *,
        request_type: global___RequestType.ValueType = ...,
        iterator_done: builtins.bool = ...,
        initialized: builtins.bool = ...,
        data: global___TrainingDataRows | None = ...,
    ) -> None: ...
    def HasField(self, field_name: typing.Literal["data", b"data", "initialized", b"initialized", "result", b"result"]) -> builtins.bool: ...
    def ClearField(self, field_name: typing.Literal["data", b"data", "initialized", b"initialized", "iterator_done", b"iterator_done", "request_type", b"request_type", "result", b"result"]) -> None: ...
    def WhichOneof(self, oneof_group: typing.Literal["result", b"result"]) -> typing.Literal["initialized", "data"] | None: ...

global___BatchTrainTestSplitResponse = BatchTrainTestSplitResponse

@typing.final
class TrainingDataRows(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    ROWS_FIELD_NUMBER: builtins.int
    @property
    def rows(self) -> google.protobuf.internal.containers.RepeatedCompositeFieldContainer[global___TrainingDataRow]: ...
    def __init__(
        self,
        *,
        rows: collections.abc.Iterable[global___TrainingDataRow] | None = ...,
    ) -> None: ...
    def ClearField(self, field_name: typing.Literal["rows", b"rows"]) -> None: ...

global___TrainingDataRows = TrainingDataRows
